import json
import time

import pytest

from labelbox import MediaType, OntologyBuilder, OntologyKind, Tool
from labelbox.orm.model import Entity
from labelbox.schema.tool_building.classification import Classification, Option
from labelbox.schema.tool_building.fact_checking_tool import (
    FactCheckingTool,
)
from labelbox.schema.tool_building.prompt_issue_tool import PromptIssueTool
from labelbox.schema.tool_building.step_reasoning_tool import (
    StepReasoningTool,
)
from labelbox.schema.tool_building.tool_type import ToolType


def test_feature_schema_is_not_archived(client, ontology):
    feature_schema_to_check = ontology.normalized["tools"][0]
    result = client.is_feature_schema_archived(
        ontology.uid, feature_schema_to_check["featureSchemaId"]
    )
    assert result is False


def test_feature_schema_is_archived(client, configured_project_with_label):
    project, _, _, label = configured_project_with_label
    ontology = project.ontology()
    feature_schema_id = ontology.normalized["tools"][0]["featureSchemaId"]
    result = client.delete_feature_schema_from_ontology(
        ontology.uid, feature_schema_id
    )
    assert result.archived is True and result.deleted is False
    assert (
        client.is_feature_schema_archived(ontology.uid, feature_schema_id)
        is True
    )


def test_is_feature_schema_archived_for_non_existing_feature_schema(
    client, ontology
):
    with pytest.raises(
        Exception, match="The specified feature schema was not in the ontology"
    ):
        client.is_feature_schema_archived(
            ontology.uid, "invalid-feature-schema-id"
        )


def test_is_feature_schema_archived_for_non_existing_ontology(client, ontology):
    feature_schema_to_unarchive = ontology.normalized["tools"][0]
    with pytest.raises(
        Exception,
        match="Resource 'Ontology' not found for params: 'invalid-ontology'",
    ):
        client.is_feature_schema_archived(
            "invalid-ontology", feature_schema_to_unarchive["featureSchemaId"]
        )


def test_delete_tool_feature_from_ontology(client, ontology):
    feature_schema_to_delete = ontology.normalized["tools"][0]
    assert len(ontology.normalized["tools"]) == 2
    result = client.delete_feature_schema_from_ontology(
        ontology.uid, feature_schema_to_delete["featureSchemaId"]
    )
    assert result.deleted is True
    assert result.archived is False
    updatedOntology = client.get_ontology(ontology.uid)
    assert len(updatedOntology.normalized["tools"]) == 1


@pytest.mark.skip(
    reason="normalized ontology contains Relationship, "
    "which is not finalized yet. introduce this back when"
    "Relationship feature is complete and we introduce"
    "a Relationship object to the ontology that we can parse"
)
def test_from_project_ontology(project) -> None:
    o = OntologyBuilder.from_project(project)
    assert o.asdict() == project.ontology().normalized


point = Tool(
    tool=Tool.Type.POINT,
    name="name",
    color="#ff0000",
)


def test_deletes_an_ontology(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized["featureSchemaId"]
    ontology = client.create_ontology_from_feature_schemas(
        name="ontology name",
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image,
    )

    assert client.delete_unused_ontology(ontology.uid) is None

    client.delete_unused_feature_schema(feature_schema_id)


def test_cant_delete_an_ontology_with_project(client):
    project = client.create_project(
        name="test project",
        media_type=MediaType.Image,
    )
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized["featureSchemaId"]
    ontology = client.create_ontology_from_feature_schemas(
        name="ontology name",
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image,
    )
    project.connect_ontology(ontology)

    with pytest.raises(
        Exception,
        match="Failed to delete the ontology, message: Cannot delete an ontology connected to a project. The ontology is connected to projects: "
        + project.uid,
    ):
        client.delete_unused_ontology(ontology.uid)

    project.delete()
    client.delete_unused_ontology(ontology.uid)
    client.delete_unused_feature_schema(feature_schema_id)


def test_inserts_a_feature_schema_at_given_position(client):
    tool1 = {"tool": "polygon", "name": "tool1", "color": "blue"}
    tool2 = {"tool": "polygon", "name": "tool2", "color": "blue"}
    ontology_normalized_json = {"tools": [tool1, tool2], "classifications": []}
    ontology = client.create_ontology(
        name="ontology",
        normalized=ontology_normalized_json,
        media_type=MediaType.Image,
    )
    created_feature_schema = client.upsert_feature_schema(point.asdict())
    client.insert_feature_schema_into_ontology(
        created_feature_schema.normalized["featureSchemaId"], ontology.uid, 1
    )
    ontology = client.get_ontology(ontology.uid)

    assert (
        ontology.normalized["tools"][1]["schemaNodeId"]
        == created_feature_schema.normalized["schemaNodeId"]
    )

    client.delete_unused_ontology(ontology.uid)


def test_moves_already_added_feature_schema_in_ontology(client):
    tool1 = {"tool": "polygon", "name": "tool1", "color": "blue"}
    ontology_normalized_json = {"tools": [tool1], "classifications": []}
    ontology = client.create_ontology(
        name="ontology",
        normalized=ontology_normalized_json,
        media_type=MediaType.Image,
    )
    created_feature_schema = client.upsert_feature_schema(point.asdict())
    feature_schema_id = created_feature_schema.normalized["featureSchemaId"]
    client.insert_feature_schema_into_ontology(
        feature_schema_id, ontology.uid, 1
    )
    ontology = client.get_ontology(ontology.uid)
    assert (
        ontology.normalized["tools"][1]["schemaNodeId"]
        == created_feature_schema.normalized["schemaNodeId"]
    )
    client.insert_feature_schema_into_ontology(
        feature_schema_id, ontology.uid, 0
    )
    ontology = client.get_ontology(ontology.uid)

    assert (
        ontology.normalized["tools"][0]["schemaNodeId"]
        == created_feature_schema.normalized["schemaNodeId"]
    )

    client.delete_unused_ontology(ontology.uid)


def test_does_not_include_used_ontologies(client):
    tool = client.upsert_feature_schema(point.asdict())
    feature_schema_id = tool.normalized["featureSchemaId"]
    ontology_with_project = client.create_ontology_from_feature_schemas(
        name="ontology name",
        feature_schema_ids=[feature_schema_id],
        media_type=MediaType.Image,
    )
    project = client.create_project(
        name="test project",
        media_type=MediaType.Image,
    )
    project.connect_ontology(ontology_with_project)
    unused_ontologies = client.get_unused_ontologies()

    assert ontology_with_project.uid not in unused_ontologies

    project.delete()
    client.delete_unused_ontology(ontology_with_project.uid)
    client.delete_unused_feature_schema(feature_schema_id)


def _get_attr_stringify_json(obj, attr):
    value = getattr(obj, attr.name)
    if attr.field_type.name.lower() == "json":
        return json.dumps(value, sort_keys=True)
    return value


@pytest.fixture
def name_for_read(rand_gen):
    yield f"test-root-schema-{rand_gen(str)}"


@pytest.fixture
def feature_schema_cat_normalized(name_for_read):
    yield {
        "tool": "polygon",
        "name": name_for_read,
        "color": "black",
        "classifications": [],
    }


@pytest.fixture
def feature_schema_for_read(client, feature_schema_cat_normalized):
    feature_schema = client.create_feature_schema(feature_schema_cat_normalized)
    yield feature_schema
    client.delete_unused_feature_schema(feature_schema.uid)


def test_feature_schema_create_read(
    client, feature_schema_for_read, name_for_read
):
    created_feature_schema = feature_schema_for_read
    queried_feature_schema = client.get_feature_schema(
        created_feature_schema.uid
    )
    for attr in Entity.FeatureSchema.fields():
        assert _get_attr_stringify_json(
            created_feature_schema, attr
        ) == _get_attr_stringify_json(queried_feature_schema, attr)

    time.sleep(3)  # Slight delay for searching
    queried_feature_schemas = list(client.get_feature_schemas(name_for_read))
    assert [
        feature_schema.name for feature_schema in queried_feature_schemas
    ] == [name_for_read]
    queried_feature_schema = queried_feature_schemas[0]

    for attr in Entity.FeatureSchema.fields():
        assert _get_attr_stringify_json(
            created_feature_schema, attr
        ) == _get_attr_stringify_json(queried_feature_schema, attr)


def test_ontology_create_read(
    client,
    rand_gen,
):
    ontology_name = f"test-ontology-{rand_gen(str)}"
    tool_name = f"test-ontology-tool-{rand_gen(str)}"
    feature_schema_cat_normalized = {
        "tool": "polygon",
        "name": tool_name,
        "color": "black",
        "classifications": [],
    }
    feature_schema = client.create_feature_schema(feature_schema_cat_normalized)
    created_ontology = client.create_ontology_from_feature_schemas(
        name=ontology_name,
        feature_schema_ids=[feature_schema.uid],
        media_type=MediaType.Image,
    )
    tool_normalized = created_ontology.normalized["tools"][0]
    for k, v in feature_schema_cat_normalized.items():
        assert tool_normalized[k] == v
    assert tool_normalized["schemaNodeId"] is not None
    assert tool_normalized["featureSchemaId"] == feature_schema.uid

    queried_ontology = client.get_ontology(created_ontology.uid)

    for attr in Entity.Ontology.fields():
        assert _get_attr_stringify_json(
            created_ontology, attr
        ) == _get_attr_stringify_json(queried_ontology, attr)

    time.sleep(3)  # Slight delay for searching
    queried_ontologies = list(client.get_ontologies(ontology_name))
    assert [ontology.name for ontology in queried_ontologies] == [ontology_name]
    queried_ontology = queried_ontologies[0]
    for attr in Entity.Ontology.fields():
        assert _get_attr_stringify_json(
            created_ontology, attr
        ) == _get_attr_stringify_json(queried_ontology, attr)


def test_unarchive_feature_schema_node(client, ontology):
    feature_schema_to_unarchive = ontology.normalized["tools"][0]
    result = client.unarchive_feature_schema_node(
        ontology.uid, feature_schema_to_unarchive["featureSchemaId"]
    )
    assert result is None


def test_unarchive_feature_schema_node_for_non_existing_feature_schema(
    client, ontology
):
    with pytest.raises(
        Exception,
        match="Failed to find feature schema node by id: invalid-feature-schema-id",
    ):
        client.unarchive_feature_schema_node(
            ontology.uid, "invalid-feature-schema-id"
        )


def test_unarchive_feature_schema_node_for_non_existing_ontology(
    client, ontology
):
    feature_schema_to_unarchive = ontology.normalized["tools"][0]
    with pytest.raises(
        Exception, match="Failed to find ontology by id: invalid-ontology"
    ):
        client.unarchive_feature_schema_node(
            "invalid-ontology", feature_schema_to_unarchive["featureSchemaId"]
        )


def test_step_reasoning_ontology(chat_evaluation_ontology):
    ontology = chat_evaluation_ontology
    step_reasoning_tool = None
    for tool in ontology.normalized["tools"]:
        if tool["tool"] == "step-reasoning":
            step_reasoning_tool = tool
            break
    assert step_reasoning_tool is not None
    assert step_reasoning_tool["definition"]["variants"] == [
        {"id": 0, "name": "Correct", "actions": []},
        {"id": 1, "name": "Neutral", "actions": []},
        {
            "id": 2,
            "name": "Incorrect",
            "actions": [
                "regenerateSteps",
                "generateAndRateAlternativeSteps",
                "rewriteStep",
                "justification",
            ],
        },
    ]
    assert step_reasoning_tool["definition"]["version"] == 1
    assert step_reasoning_tool["schemaNodeId"] is not None
    assert step_reasoning_tool["featureSchemaId"] is not None

    step_reasoning_tool = None
    for tool in ontology.tools():
        if isinstance(tool, StepReasoningTool):
            step_reasoning_tool = tool
            break
    assert step_reasoning_tool is not None

    assert step_reasoning_tool.definition.asdict() == {
        "title": "step reasoning",
        "value": "step_reasoning",
        "variants": [
            {
                "id": 0,
                "name": "Correct",
                "actions": [],
            },
            {
                "id": 1,
                "name": "Neutral",
                "actions": [],
            },
            {
                "id": 2,
                "name": "Incorrect",
                "actions": [
                    "regenerateSteps",
                    "generateAndRateAlternativeSteps",
                    "rewriteStep",
                    "justification",
                ],
            },
        ],
        "version": 1,
    }


def test_fact_checking_ontology(chat_evaluation_ontology):
    ontology = chat_evaluation_ontology
    fact_checking = None
    for tool in ontology.normalized["tools"]:
        if tool["tool"] == "fact-checking":
            fact_checking = tool
            break
    assert fact_checking is not None
    assert fact_checking["definition"]["variants"] == [
        {"id": 0, "name": "Accurate", "actions": ["justification"]},
        {"id": 1, "name": "Inaccurate", "actions": ["justification"]},
        {"id": 2, "name": "Disputed", "actions": ["justification"]},
        {"id": 3, "name": "Unsupported", "actions": []},
        {
            "id": 4,
            "name": "Can't confidently assess",
            "actions": [],
        },
        {
            "id": 5,
            "name": "No factual information",
            "actions": [],
        },
    ]
    assert fact_checking["definition"]["version"] == 1
    assert fact_checking["schemaNodeId"] is not None
    assert fact_checking["featureSchemaId"] is not None

    fact_checking = None
    for tool in ontology.tools():
        if isinstance(tool, FactCheckingTool):
            fact_checking = tool
            break
    assert fact_checking is not None

    assert fact_checking.definition.asdict() == {
        "title": "fact checking",
        "value": "fact_checking",
        "variants": [
            {"id": 0, "name": "Accurate", "actions": ["justification"]},
            {"id": 1, "name": "Inaccurate", "actions": ["justification"]},
            {"id": 2, "name": "Disputed", "actions": ["justification"]},
            {"id": 3, "name": "Unsupported", "actions": []},
            {
                "id": 4,
                "name": "Can't confidently assess",
                "actions": [],
            },
            {
                "id": 5,
                "name": "No factual information",
                "actions": [],
            },
        ],
        "version": 1,
    }


def test_prompt_issue_ontology(chat_evaluation_ontology):
    ontology = chat_evaluation_ontology
    prompt_issue = None
    for tool in ontology.normalized["tools"]:
        if tool["tool"] == "prompt-issue":
            prompt_issue = tool
            break
    assert prompt_issue is not None

    assert prompt_issue["definition"] == {
        "title": "prompt issue",
        "value": "prompt_issue",
        "color": "#ff00ff",
    }
    assert prompt_issue["schemaNodeId"] is not None
    assert prompt_issue["featureSchemaId"] is not None
    assert len(prompt_issue["classifications"]) == 1

    prompt_issue_tool = None
    for tool in ontology.tools():
        if isinstance(tool, PromptIssueTool):
            prompt_issue_tool = tool
            break
    assert prompt_issue_tool is not None
    # Assertions
    assert prompt_issue_tool.name == "prompt issue"
    assert prompt_issue_tool.type == ToolType.PROMPT_ISSUE
    assert prompt_issue_tool.schema_id is not None
    assert prompt_issue_tool.feature_schema_id is not None

    # Check classifications
    assert len(prompt_issue_tool.classifications) == 1
    classification = prompt_issue_tool.classifications[0]
    assert classification.class_type == Classification.Type.CHECKLIST
    assert len(classification.options) == 3  # Check number of options


def test_invalid_prompt_issue_ontology(client):
    tool = PromptIssueTool(name="Prompt Issue Tool")

    option1 = Option(value="value")
    radio_class = Classification(
        class_type=Classification.Type.RADIO,
        name="radio-class",
        options=[option1],
    )
    text_class = Classification(
        class_type=Classification.Type.TEXT, name="text-class"
    )

    tool.classifications.append(radio_class)
    tool.classifications.append(text_class)

    builder = OntologyBuilder(
        tools=[tool],
    )
    with pytest.raises(
        ValueError, match="Classifications for Prompt Issue Tool are invalid"
    ):
        client.create_ontology(
            name="plt-1710",
            media_type=MediaType.Conversational,
            ontology_kind=OntologyKind.ModelEvaluation,
            normalized=builder.asdict(),
        )



from typing import List


from labelbox import Project, Dataset


def test_create_batches(project: Project, big_dataset_data_row_ids: List[str]):
    task = project.create_batches(
        "test-batch", big_dataset_data_row_ids, priority=3
    )

    task.wait_till_done()
    assert task.errors() is None
    batches = task.result()

    assert len(batches) == 1
    assert batches[0].name == "test-batch0000"
    assert batches[0].size == len(big_dataset_data_row_ids)


def test_create_batches_from_dataset(project: Project, big_dataset: Dataset):
    export_task = big_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    project._wait_until_data_rows_are_processed(data_rows, [], 300)

    task = project.create_batches_from_dataset(
        "test-batch", big_dataset.uid, priority=3
    )

    task.wait_till_done()
    assert task.errors() is None
    batches = task.result()

    assert len(batches) == 1
    assert batches[0].name == "test-batch0000"
    assert batches[0].size == len(data_rows)



import pytest
from labelbox.utils import (
    format_iso_datetime,
    format_iso_from_string,
    sentence_case,
)


@pytest.mark.parametrize(
    "datetime_str, expected_datetime_str",
    [
        ("2011-11-04T00:05:23Z", "2011-11-04T00:05:23Z"),
        ("2011-11-04T00:05:23+00:00", "2011-11-04T00:05:23Z"),
        ("2011-11-04T00:05:23+05:00", "2011-11-03T19:05:23Z"),
        ("2011-11-04T00:05:23", "2011-11-04T00:05:23Z"),
    ],
)
def test_datetime_parsing(datetime_str, expected_datetime_str):
    # NOTE I would normally not take 'expected' using another function from sdk code, but in this case this is exactly the usage in _validate_parse_datetime
    assert (
        format_iso_datetime(format_iso_from_string(datetime_str))
        == expected_datetime_str
    )


@pytest.mark.parametrize(
    "str, expected_str",
    [
        ("AUDIO", "Audio"),
        ("LLM_PROMPT_RESPONSE_CREATION", "Llm prompt response creation"),
    ],
)
def test_sentence_case(str, expected_str):
    assert sentence_case(str) == expected_str



from labelbox.data.annotation_types.classification.classification import Text


def test_text():
    text_entity = Text(answer="good job")
    assert text_entity.answer == "good job"


def test_text_confidence():
    text_entity = Text(answer="good job", confidence=0.5)
    assert text_entity.answer == "good job"
    assert text_entity.confidence == 0.5



from labelbox import UniqueIds, Project, Ontology, Client
from labelbox.schema.conflict_resolution_strategy import (
    ConflictResolutionStrategy,
)
from typing import List


def test_send_to_annotate_include_annotations(
    client: Client,
    configured_batch_project_with_label: Project,
    project_pack: List[Project],
    ontology: Ontology,
):
    [source_project, _, data_row, _] = configured_batch_project_with_label
    destination_project: Project = project_pack[0]

    src_ontology = source_project.ontology()
    destination_project.connect_ontology(ontology)

    # build an ontology mapping using the top level tools
    src_feature_schema_ids = list(
        tool.feature_schema_id for tool in src_ontology.tools()
    )
    dest_ontology = destination_project.ontology()
    dest_feature_schema_ids = list(
        tool.feature_schema_id for tool in dest_ontology.tools()
    )
    # create a dictionary of feature schema id to itself
    ontology_mapping = dict(
        zip(src_feature_schema_ids, dest_feature_schema_ids)
    )

    try:
        queues = destination_project.task_queues()
        initial_review_task = next(
            q for q in queues if q.name == "Initial review task"
        )

        # Send the data row to the new project
        task = client.send_to_annotate_from_catalog(
            destination_project_id=destination_project.uid,
            task_queue_id=initial_review_task.uid,
            batch_name="test-batch",
            data_rows=UniqueIds([data_row.uid]),
            params={
                "source_project_id": source_project.uid,
                "annotations_ontology_mapping": ontology_mapping,
                "override_existing_annotations_rule": ConflictResolutionStrategy.OverrideWithAnnotations,
            },
        )

        task.wait_till_done()

        # Check that the data row was sent to the new project
        destination_batches = list(destination_project.batches())
        assert len(destination_batches) == 1

        export_task = destination_project.export()
        export_task.wait_till_done()
        stream = export_task.get_buffered_stream()

        destination_data_rows = [dr.json["data_row"]["id"] for dr in stream]
        assert len(destination_data_rows) == 1
        assert destination_data_rows[0] == data_row.uid

        # Verify annotations were copied into the destination project
        destination_project_labels = list(destination_project.labels())
        assert len(destination_project_labels) == 1
    finally:
        destination_project.delete()



import os
import time

import pytest

from labelbox import Client, Label


def test_labels(configured_project_with_label):
    project, _, data_row, label = configured_project_with_label

    assert list(project.labels()) == [label]
    assert list(data_row.labels()) == [label]

    assert label.project() == project
    assert label.data_row() == data_row
    assert label.created_by() == label.client.get_user()

    label.delete()

    # TODO: Added sleep to account for ES from catching up to deletion.
    # Need a better way to query labels in `project.labels()`, because currently,
    # it intermittently takes too long to sync, causing flaky SDK tests
    time.sleep(5)

    assert list(project.labels()) == []
    assert list(data_row.labels()) == []


# TODO: Skipping this test in staging due to label not updating
@pytest.mark.skipif(
    condition=os.environ["LABELBOX_TEST_ENVIRON"] == "onprem"
    or os.environ["LABELBOX_TEST_ENVIRON"] == "staging"
    or os.environ["LABELBOX_TEST_ENVIRON"] == "local"
    or os.environ["LABELBOX_TEST_ENVIRON"] == "custom",
    reason="does not work for onprem",
)
def test_label_update(configured_project_with_label):
    _, _, _, label = configured_project_with_label
    label.update(label="something else")
    assert label.label == "something else"


def test_label_filter_order(configured_project_with_label, label_helpers):
    project, _, _, label = configured_project_with_label

    project.create_label()
    label_helpers.wait_for_labels(project, 2)

    list_asc = list(project.labels(order_by=Label.created_at.asc))
    list_desc = list(project.labels(order_by=Label.created_at.desc))
    assert list_asc == list_desc[::-1]


def test_label_bulk_deletion(configured_project_with_label):
    project, _, _, _ = configured_project_with_label

    for _ in range(2):
        # only run twice, already have one label in the fixture
        project.create_label()
    labels = project.labels()
    l1 = next(labels)
    l2 = next(labels)
    l3 = next(labels)

    assert set(project.labels()) == {l1, l2, l3}

    Label.bulk_delete([l1, l3])

    # TODO: the sdk client should really abstract all these timing issues away
    # but for now bulk deletes take enough time that this test is flaky
    # add sleep here to avoid that flake
    time.sleep(5)

    assert set(project.labels()) == {l2}


def test_upsert_label_scores(configured_project_with_label, client: Client):
    project, _, _, _ = configured_project_with_label

    label = next(project.labels())

    scores = client.upsert_label_feedback(
        label_id=label.uid,
        feedback="That's a great label!",
        scores={"overall": 5},
    )
    assert len(scores) == 1
    assert scores[0].score == 5



from dataclasses import dataclass

from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field
from labelbox.schema.role import ProjectRole, format_role


@dataclass
class InviteLimit:
    """
    remaining (int): Number of invites remaining in the org
    used (int): Number of invites used in the org
    limit (int): Maximum number of invites available to the org
    """

    remaining: int
    used: int
    limit: int


class Invite(DbObject):
    """
    An object representing a user invite
    """

    created_at = Field.DateTime("created_at")
    organization_role_name = Field.String("organization_role_name")
    email = Field.String("email", "inviteeEmail")

    def __init__(self, client, invite_response):
        project_roles = invite_response.pop("projectInvites", [])
        super().__init__(client, invite_response)

        self.project_roles = [
            ProjectRole(
                project=client.get_project(r["projectId"]),
                role=client.get_roles()[format_role(r["projectRoleName"])],
            )
            for r in project_roles
        ]



import pytest

from labelbox.schema.tool_building.step_reasoning_tool import (
    StepReasoningTool,
)


def test_validations():
    with pytest.raises(ValueError):
        StepReasoningTool(name="")


def test_step_reasoning_as_dict_default():
    tool = StepReasoningTool(name="step reasoning")
    assert tool.definition.variants[2]._available_actions == {
        "regenerateSteps",
        "generateAndRateAlternativeSteps",
        "rewriteStep",
        "justification",
    }

    assert tool.asdict() == {
        "tool": "step-reasoning",
        "name": "step reasoning",
        "required": False,
        "schemaNodeId": None,
        "featureSchemaId": None,
        "color": None,
        "definition": {
            "variants": [
                {"id": 0, "name": "Correct", "actions": []},
                {"id": 1, "name": "Neutral", "actions": []},
                {
                    "id": 2,
                    "name": "Incorrect",
                    "actions": [
                        "regenerateSteps",
                        "generateAndRateAlternativeSteps",
                        "rewriteStep",
                        "justification",
                    ],
                },
            ],
            "version": 1,
        },
    }


def test_from_dict():
    dict = {
        "schemaNodeId": "cm3pdkupv0ah8070h2ujo74th",
        "featureSchemaId": "cm3pdkupv0ah7070hg7svdeeo",
        "required": False,
        "name": "step reasoning",
        "tool": "step-reasoning",
        "definition": {
            "variants": [
                {"id": 0, "name": "Correct", "actions": []},
                {"id": 1, "name": "Neutral", "actions": []},
                {
                    "id": 2,
                    "name": "Incorrect",
                    "actions": [
                        "regenerateSteps",
                        "generateAndRateAlternativeSteps",
                        "rewriteStep",
                        "justification",
                    ],
                },
            ],
            "version": 1,
            "title": "step reasoning",
            "value": "step_reasoning",
            "color": "#ff0000",
        },
        "color": "#ff0000",
        "archived": 0,
        "classifications": [],
        "kind": "StepReasoning",
    }
    tool = StepReasoningTool.from_dict(dict)
    assert tool.definition.variants[2]._available_actions == {
        "generateAndRateAlternativeSteps",
        "justification",
        "rewriteStep",
        "regenerateSteps",
    }



from .scalar import ScalarMetric, ScalarMetricAggregation, ScalarMetricValue
from .confusion_matrix import (
    ConfusionMatrixMetric,
    ConfusionMatrixAggregation,
    ConfusionMatrixMetricValue,
)



import pytest
from lbox.exceptions import ResourceNotFoundError


def test_create_model_config(client, valid_model_id):
    model_config = client.create_model_config(
        "model_config", valid_model_id, {"param": "value"}
    )
    assert model_config.inference_params["param"] == "value"
    assert model_config.name == "model_config"
    assert model_config.model_id == valid_model_id


def test_delete_model_config(client, valid_model_id):
    model_config_id = client.create_model_config(
        "model_config", valid_model_id, {"param": "value"}
    )
    assert client.delete_model_config(model_config_id.uid)


def test_delete_nonexistant_model_config(client):
    with pytest.raises(ResourceNotFoundError):
        client.delete_model_config("invalid_model_id")



from .line import Line
from .point import Point
from .mask import Mask
from .polygon import Polygon
from .rectangle import Rectangle
from .rectangle import DocumentRectangle
from .rectangle import RectangleUnit
from .geometry import Geometry



from labelbox.data.annotation_types import Label, VideoObjectAnnotation
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.data.annotation_types.geometry import Rectangle, Point
from labelbox.data.annotation_types.data import GenericDataRowData


def video_bbox_label():
    return Label(
        uid="cl1z52xwh00050fhcmfgczqvn",
        data=GenericDataRowData(
            uid="cklr9mr4m5iao0rb6cvxu4qbn",
            url="https://storage.labelbox.com/ckcz6bubudyfi0855o1dt1g9s%2F26403a22-604a-a38c-eeff-c2ed481fb40a-cat.mp4?Expires=1651677421050&KeyName=labelbox-assets-key-3&Signature=vF7gMyfHzgZdfbB8BHgd88Ws-Ms",
        ),
        annotations=[
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                    "uuid": "b24e672b-8f79-4d96-bf5e-b552ca0820d5",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=46.0),
                    end=Point(extra={}, x=454.0, y=295.0),
                ),
                classifications=[],
                frame=1,
                keyframe=True,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=42.5),
                    end=Point(extra={}, x=427.25, y=308.25),
                ),
                classifications=[],
                frame=2,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=39.0),
                    end=Point(extra={}, x=400.5, y=321.5),
                ),
                classifications=[],
                frame=3,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=35.5),
                    end=Point(extra={}, x=373.75, y=334.75),
                ),
                classifications=[],
                frame=4,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=32.0),
                    end=Point(extra={}, x=347.0, y=348.0),
                ),
                classifications=[],
                frame=5,
                keyframe=True,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=132.0),
                    end=Point(extra={}, x=283.0, y=348.0),
                ),
                classifications=[],
                frame=9,
                keyframe=True,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=122.333),
                    end=Point(extra={}, x=295.5, y=348.0),
                ),
                classifications=[],
                frame=10,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=112.667),
                    end=Point(extra={}, x=308.0, y=348.0),
                ),
                classifications=[],
                frame=11,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=103.0),
                    end=Point(extra={}, x=320.5, y=348.0),
                ),
                classifications=[],
                frame=12,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=93.333),
                    end=Point(extra={}, x=333.0, y=348.0),
                ),
                classifications=[],
                frame=13,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=83.667),
                    end=Point(extra={}, x=345.5, y=348.0),
                ),
                classifications=[],
                frame=14,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=74.0),
                    end=Point(extra={}, x=358.0, y=348.0),
                ),
                classifications=[],
                frame=15,
                keyframe=True,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=66.833),
                    end=Point(extra={}, x=387.333, y=348.0),
                ),
                classifications=[],
                frame=16,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=59.667),
                    end=Point(extra={}, x=416.667, y=348.0),
                ),
                classifications=[],
                frame=17,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=52.5),
                    end=Point(extra={}, x=446.0, y=348.0),
                ),
                classifications=[],
                frame=18,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=45.333),
                    end=Point(extra={}, x=475.333, y=348.0),
                ),
                classifications=[],
                frame=19,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=38.167),
                    end=Point(extra={}, x=504.667, y=348.0),
                ),
                classifications=[],
                frame=20,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=31.0),
                    end=Point(extra={}, x=534.0, y=348.0),
                ),
                classifications=[],
                frame=21,
                keyframe=True,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=29.5),
                    end=Point(extra={}, x=543.0, y=348.0),
                ),
                classifications=[],
                frame=22,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=28.0),
                    end=Point(extra={}, x=552.0, y=348.0),
                ),
                classifications=[],
                frame=23,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=26.5),
                    end=Point(extra={}, x=561.0, y=348.0),
                ),
                classifications=[],
                frame=24,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=25.0),
                    end=Point(extra={}, x=570.0, y=348.0),
                ),
                classifications=[],
                frame=25,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=23.5),
                    end=Point(extra={}, x=579.0, y=348.0),
                ),
                classifications=[],
                frame=26,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=22.0),
                    end=Point(extra={}, x=588.0, y=348.0),
                ),
                classifications=[],
                frame=27,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=20.5),
                    end=Point(extra={}, x=597.0, y=348.0),
                ),
                classifications=[],
                frame=28,
                keyframe=False,
            ),
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=19.0),
                    end=Point(extra={}, x=606.0, y=348.0),
                ),
                classifications=[],
                frame=29,
                keyframe=True,
            ),
        ],
        extra={
            "Created By": "jtso@labelbox.com",
            "Project Name": "Pictor Video",
            "Created At": "2022-04-14T15:11:19.000Z",
            "Updated At": "2022-04-14T15:11:21.064Z",
            "Seconds to Label": 0.0,
            "Agreement": -1.0,
            "Benchmark Agreement": -1.0,
            "Benchmark ID": None,
            "Dataset Name": "cat",
            "Reviews": [],
            "View Label": "https://editor.labelbox.com?project=ckz38nsfd0lzq109bhq73est1&label=cl1z52xwh00050fhcmfgczqvn",
            "Has Open Issues": 0.0,
            "Skipped": False,
            "media_type": "video",
            "Data Split": None,
        },
    )


def video_serialized_bbox_label():
    return {
        "uuid": "b24e672b-8f79-4d96-bf5e-b552ca0820d5",
        "dataRow": {"id": "cklr9mr4m5iao0rb6cvxu4qbn"},
        "schemaId": "ckz38ofop0mci0z9i9w3aa9o4",
        "name": "bbox toy",
        "classifications": [],
        "segments": [
            {
                "keyframes": [
                    {
                        "frame": 1,
                        "bbox": {
                            "top": 46.0,
                            "left": 70.0,
                            "height": 249.0,
                            "width": 384.0,
                        },
                        "classifications": [],
                    },
                    {
                        "frame": 5,
                        "bbox": {
                            "top": 32.0,
                            "left": 70.0,
                            "height": 316.0,
                            "width": 277.0,
                        },
                        "classifications": [],
                    },
                ]
            },
            {
                "keyframes": [
                    {
                        "frame": 9,
                        "bbox": {
                            "top": 132.0,
                            "left": 70.0,
                            "height": 216.0,
                            "width": 213.0,
                        },
                        "classifications": [],
                    },
                    {
                        "frame": 15,
                        "bbox": {
                            "top": 74.0,
                            "left": 70.0,
                            "height": 274.0,
                            "width": 288.0,
                        },
                        "classifications": [],
                    },
                    {
                        "frame": 21,
                        "bbox": {
                            "top": 31.0,
                            "left": 70.0,
                            "height": 317.0,
                            "width": 464.0,
                        },
                        "classifications": [],
                    },
                    {
                        "frame": 29,
                        "bbox": {
                            "top": 19.0,
                            "left": 70.0,
                            "height": 329.0,
                            "width": 536.0,
                        },
                        "classifications": [],
                    },
                ]
            },
        ],
    }


def test_serialize_video_objects():
    label = video_bbox_label()
    serialized_labels = NDJsonConverter.serialize([label])
    label = next(serialized_labels)

    assert label == video_serialized_bbox_label()



import pytest
from lbox.exceptions import ConfidenceNotSupportedException
from pydantic import ValidationError

from labelbox.data.annotation_types import (
    ClassificationAnnotation,
    Line,
    ObjectAnnotation,
    Point,
    Text,
    TextEntity,
)
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.video import (
    VideoClassificationAnnotation,
    VideoObjectAnnotation,
)


def test_annotation():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])
    classification = Text(answer="1234")

    annotation = ObjectAnnotation(
        value=line,
        name=name,
    )
    assert annotation.value.points[0].model_dump() == {
        "extra": {},
        "x": 1.0,
        "y": 2.0,
    }
    assert annotation.name == name

    # Check ner
    ObjectAnnotation(
        value=TextEntity(start=10, end=12),
        name=name,
    )

    # Check classification
    ClassificationAnnotation(
        value=classification,
        name=name,
    )

    # Invalid subclass
    with pytest.raises(ValidationError):
        ObjectAnnotation(
            value=line,
            name=name,
            classifications=[line],
        )

    subclass = ClassificationAnnotation(value=classification, name=name)

    ObjectAnnotation(
        value=line,
        name=name,
        classifications=[subclass],
    )


def test_video_annotations():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])

    # Wrong type
    with pytest.raises(ValidationError):
        VideoClassificationAnnotation(value=line, name=name, frame=1)

    # Missing frames
    with pytest.raises(ValidationError):
        VideoClassificationAnnotation(value=line, name=name)

    VideoObjectAnnotation(value=line, name=name, keyframe=True, frame=2)


def test_confidence_for_video_is_not_supported():
    with pytest.raises(ConfidenceNotSupportedException):
        (
            VideoObjectAnnotation(
                name="bbox toy",
                feature_schema_id="ckz38ofop0mci0z9i9w3aa9o4",
                extra={
                    "value": "bbox_toy",
                    "instanceURI": None,
                    "color": "#1CE6FF",
                    "feature_id": "cl1z52xw700000fhcayaqy0ev",
                },
                value=Rectangle(
                    extra={},
                    start=Point(extra={}, x=70.0, y=26.5),
                    end=Point(extra={}, x=561.0, y=348.0),
                ),
                classifications=[],
                frame=24,
                keyframe=False,
                confidence=0.3434,
            ),
        )


def test_confidence_value_range_validation():
    name = "line_feature"
    line = Line(points=[Point(x=1, y=2), Point(x=2, y=2)])

    with pytest.raises(ValueError) as e:
        ObjectAnnotation(value=line, name=name, confidence=14)



from labelbox import ExportTask, StreamType


class TestExportModelRun:
    def test_export(self, model_run_with_data_rows):
        model_run, labels = model_run_with_data_rows
        label_ids = [label.uid for label in labels]
        expected_data_rows = list(model_run.model_run_data_rows())

        task_name = "TestExportModelRun:test_export"
        params = {"media_attributes": True, "predictions": True}
        export_task = model_run.export(task_name, params=params)
        assert export_task.name == task_name
        export_task.wait_till_done()

        assert export_task.status == "COMPLETE"
        assert isinstance(export_task, ExportTask)
        assert export_task.has_result()
        assert export_task.has_errors() is False
        assert (
            export_task.get_total_file_size(stream_type=StreamType.RESULT) > 0
        )
        assert export_task.get_total_lines(
            stream_type=StreamType.RESULT
        ) == len(expected_data_rows)

        for data in export_task.get_buffered_stream():
            obj = data.json
            assert (
                "media_attributes" in obj
                and obj["media_attributes"] is not None
            )
            exported_model_run = obj["experiments"][model_run.model_id]["runs"][
                model_run.uid
            ]
            task_label_ids_set = set(
                map(lambda label: label["id"], exported_model_run["labels"])
            )
            task_prediction_ids_set = set(
                map(
                    lambda prediction: prediction["id"],
                    exported_model_run["predictions"],
                )
            )
            for label_id in task_label_ids_set:
                assert label_id in label_ids
            for prediction_id in task_prediction_ids_set:
                assert prediction_id in label_ids



from typing import TYPE_CHECKING

from labelbox.orm import query
from labelbox.orm.db_object import DbObject, Updateable, BulkDeletable
from labelbox.orm.model import Entity, Field, Relationship

if TYPE_CHECKING:
    from labelbox import Benchmark, Review
""" Client-side object type definitions. """


class Label(DbObject, Updateable, BulkDeletable):
    """Label represents an assessment on a DataRow. For example one label could
    contain 100 bounding boxes (annotations).

    Attributes:
        label (str)
        seconds_to_label (float)
        agreement (float)
        benchmark_agreement (float)
        is_benchmark_reference (bool)

        project (Relationship): `ToOne` relationship to Project
        data_row (Relationship): `ToOne` relationship to DataRow
        reviews (Relationship): `ToMany` relationship to Review
        created_by (Relationship): `ToOne` relationship to User

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reviews.supports_filtering = False

    label = Field.String("label")
    seconds_to_label = Field.Float("seconds_to_label")
    agreement = Field.Float("agreement")
    benchmark_agreement = Field.Float("benchmark_agreement")
    is_benchmark_reference = Field.Boolean("is_benchmark_reference")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")

    project = Relationship.ToOne("Project")
    data_row = Relationship.ToOne("DataRow")
    reviews = Relationship.ToMany("Review", False)
    created_by = Relationship.ToOne("User", False, "created_by")

    @staticmethod
    def bulk_delete(labels) -> None:
        """Deletes all the given Labels.

        Args:
            labels (list of Label): The Labels to delete.
        """
        BulkDeletable._bulk_delete(labels, False)

    def create_review(self, **kwargs) -> "Review":
        """Creates a Review for this label.

        Args:
            **kwargs: Review attributes. At a minimum, a `Review.score` field value must be provided.
        """
        kwargs[Entity.Review.label.name] = self
        kwargs[Entity.Review.project.name] = self.project()
        return self.client._create(Entity.Review, kwargs)

    def create_benchmark(self) -> "Benchmark":
        """Creates a Benchmark for this Label.

        Returns:
            The newly created Benchmark.
        """
        label_id_param = "labelId"
        query_str = """mutation CreateBenchmarkPyApi($%s: ID!) {
            createBenchmark(data: {labelId: $%s}) {%s}} """ % (
            label_id_param,
            label_id_param,
            query.results_query_part(Entity.Benchmark),
        )
        res = self.client.execute(query_str, {label_id_param: self.uid})
        return Entity.Benchmark(self.client, res["createBenchmark"])



import uuid
from datetime import datetime, timezone

import pytest
from lbox.exceptions import MalformedQueryException

from labelbox import Client, DataRow, DataRowMetadataOntology, Dataset
from labelbox.schema.data_row_metadata import (
    DataRowMetadata,
    DataRowMetadataField,
    DataRowMetadataKind,
    DeleteDataRowMetadata,
)
from labelbox.schema.identifiable import GlobalKey, UniqueId

INVALID_SCHEMA_ID = "1" * 25
FAKE_SCHEMA_ID = "0" * 25
FAKE_DATAROW_ID = "D" * 25
SPLIT_SCHEMA_ID = "cko8sbczn0002h2dkdaxb5kal"
TRAIN_SPLIT_ID = "cko8sbscr0003h2dk04w86hof"
TEST_SPLIT_ID = "cko8scbz70005h2dkastwhgqt"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"
CAPTURE_DT_SCHEMA_ID = "cko8sdzv70006h2dk8jg64zvb"
CUSTOM_TEXT_SCHEMA_NAME = "custom_text"

FAKE_NUMBER_FIELD = {
    "id": FAKE_SCHEMA_ID,
    "name": "number",
    "kind": "CustomMetadataNumber",
    "reserved": False,
}


@pytest.fixture
def mdo(client: Client):
    mdo = client.get_data_row_metadata_ontology()
    try:
        mdo.create_schema(CUSTOM_TEXT_SCHEMA_NAME, DataRowMetadataKind.string)
    except MalformedQueryException:
        # Do nothing if already exists
        pass
    mdo._raw_ontology = mdo._get_ontology()
    mdo._raw_ontology.append(FAKE_NUMBER_FIELD)
    mdo._build_ontology()
    yield mdo


@pytest.fixture
def big_dataset(dataset: Dataset, image_url):
    task = dataset.create_data_rows(
        [
            {
                "row_data": image_url,
                "external_id": "my-image",
                "global_key": str(uuid.uuid4()),
            },
        ]
        * 5
    )
    task.wait_till_done()

    yield dataset


def make_metadata(dr_id: str = None, gk: str = None) -> DataRowMetadata:
    msg = "A message"
    time = datetime.now(timezone.utc)

    metadata = DataRowMetadata(
        global_key=gk,
        data_row_id=dr_id,
        fields=[
            DataRowMetadataField(
                schema_id=SPLIT_SCHEMA_ID, value=TEST_SPLIT_ID
            ),
            DataRowMetadataField(schema_id=CAPTURE_DT_SCHEMA_ID, value=time),
            DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value=msg),
        ],
    )
    return metadata


def make_named_metadata(dr_id) -> DataRowMetadata:
    msg = "A message"
    time = datetime.now(timezone.utc)

    metadata = DataRowMetadata(
        data_row_id=dr_id,
        fields=[
            DataRowMetadataField(name="split", value=TEST_SPLIT_ID),
            DataRowMetadataField(name="captureDateTime", value=time),
            DataRowMetadataField(name=CUSTOM_TEXT_SCHEMA_NAME, value=msg),
        ],
    )
    return metadata


@pytest.mark.order(1)
def test_bulk_delete_datarow_metadata(data_row, mdo):
    """test bulk deletes for all fields"""
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    assert len(mdo.bulk_export([data_row.uid])[0].fields)
    upload_ids = [m.schema_id for m in metadata.fields[:-2]]
    mdo.bulk_delete(
        [DeleteDataRowMetadata(data_row_id=data_row.uid, fields=upload_ids)]
    )
    remaining_ids = set(
        [f.schema_id for f in mdo.bulk_export([data_row.uid])[0].fields]
    )
    assert not len(remaining_ids.intersection(set(upload_ids)))


@pytest.fixture
def data_row_unique_id(data_row):
    return UniqueId(data_row.uid)


@pytest.fixture
def data_row_global_key(data_row):
    return GlobalKey(data_row.global_key)


@pytest.fixture
def data_row_id_as_str(data_row):
    return data_row.uid


@pytest.mark.parametrize(
    "data_row_for_delete",
    ["data_row_id_as_str", "data_row_unique_id", "data_row_global_key"],
)
def test_bulk_delete_datarow_metadata(
    data_row_for_delete, data_row, mdo, request
):
    """test bulk deletes for all fields"""
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    assert len(mdo.bulk_export([data_row.uid])[0].fields)
    upload_ids = [m.schema_id for m in metadata.fields[:-2]]
    mdo.bulk_delete(
        [
            DeleteDataRowMetadata(
                data_row_id=request.getfixturevalue(data_row_for_delete),
                fields=upload_ids,
            )
        ]
    )
    remaining_ids = set(
        [f.schema_id for f in mdo.bulk_export([data_row.uid])[0].fields]
    )
    assert not len(remaining_ids.intersection(set(upload_ids)))


@pytest.mark.parametrize(
    "data_row_for_delete",
    ["data_row_id_as_str", "data_row_unique_id", "data_row_global_key"],
)
def test_bulk_partial_delete_datarow_metadata(
    data_row_for_delete, data_row, mdo, request
):
    """Delete a single from metadata"""
    n_fields = len(mdo.bulk_export([data_row.uid])[0].fields)
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])

    assert len(mdo.bulk_export([data_row.uid])[0].fields) == (
        n_fields + len(metadata.fields)
    )

    mdo.bulk_delete(
        [
            DeleteDataRowMetadata(
                data_row_id=request.getfixturevalue(data_row_for_delete),
                fields=[TEXT_SCHEMA_ID],
            )
        ]
    )
    fields = [f for f in mdo.bulk_export([data_row.uid])[0].fields]
    assert len(fields) == (len(metadata.fields) - 1)


@pytest.fixture
def data_row_unique_ids(big_dataset):
    deletes = []
    data_row_ids = [dr.uid for dr in big_dataset.data_rows()]

    for data_row_id in data_row_ids:
        deletes.append(
            DeleteDataRowMetadata(
                data_row_id=UniqueId(data_row_id),
                fields=[SPLIT_SCHEMA_ID, CAPTURE_DT_SCHEMA_ID],
            )
        )
    return deletes


@pytest.fixture
def data_row_ids_as_str(big_dataset):
    deletes = []
    data_row_ids = [dr.uid for dr in big_dataset.data_rows()]

    for data_row_id in data_row_ids:
        deletes.append(
            DeleteDataRowMetadata(
                data_row_id=data_row_id,
                fields=[SPLIT_SCHEMA_ID, CAPTURE_DT_SCHEMA_ID],
            )
        )
    return deletes


@pytest.fixture
def data_row_global_keys(big_dataset):
    deletes = []
    global_keys = [dr.global_key for dr in big_dataset.data_rows()]

    for data_row_id in global_keys:
        deletes.append(
            DeleteDataRowMetadata(
                data_row_id=GlobalKey(data_row_id),
                fields=[SPLIT_SCHEMA_ID, CAPTURE_DT_SCHEMA_ID],
            )
        )
    return deletes


@pytest.mark.parametrize(
    "data_rows_for_delete",
    ["data_row_ids_as_str", "data_row_unique_ids", "data_row_global_keys"],
)
def test_large_bulk_delete_datarow_metadata(
    data_rows_for_delete, big_dataset, mdo, request
):
    metadata = []
    data_row_ids = [dr.uid for dr in big_dataset.data_rows()]
    for data_row_id in data_row_ids:
        metadata.append(
            DataRowMetadata(
                data_row_id=data_row_id,
                fields=[
                    DataRowMetadataField(
                        schema_id=SPLIT_SCHEMA_ID, value=TEST_SPLIT_ID
                    ),
                    DataRowMetadataField(
                        schema_id=TEXT_SCHEMA_ID, value="test-message"
                    ),
                ],
            )
        )
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    deletes = request.getfixturevalue(data_rows_for_delete)
    errors = mdo.bulk_delete(deletes)

    assert len(errors) == len(data_row_ids)
    for error in errors:
        assert error.fields == [CAPTURE_DT_SCHEMA_ID]
        assert error.error == "Schema did not exist"

    for data_row_id in data_row_ids:
        fields = [f for f in mdo.bulk_export([data_row_id])[0].fields]
        assert len(fields) == 1, fields
        assert SPLIT_SCHEMA_ID not in [field.schema_id for field in fields]


@pytest.mark.parametrize(
    "data_row_for_delete",
    ["data_row_id_as_str", "data_row_unique_id", "data_row_global_key"],
)
def test_bulk_delete_datarow_enum_metadata(
    data_row_for_delete,
    data_row: DataRow,
    mdo: DataRowMetadataOntology,
    request,
):
    """test bulk deletes for non non fields"""
    metadata = make_metadata(data_row.uid)
    metadata.fields = [
        m for m in metadata.fields if m.schema_id == SPLIT_SCHEMA_ID
    ]
    mdo.bulk_upsert([metadata])

    exported = mdo.bulk_export([data_row.uid])[0].fields
    assert len(exported) == len(
        set(
            [x.schema_id for x in metadata.fields]
            + [x.schema_id for x in exported]
        )
    )

    mdo.bulk_delete(
        [
            DeleteDataRowMetadata(
                data_row_id=request.getfixturevalue(data_row_for_delete),
                fields=[SPLIT_SCHEMA_ID],
            )
        ]
    )
    exported = mdo.bulk_export([data_row.uid])[0].fields
    assert len(exported) == 0


@pytest.mark.parametrize(
    "data_row_for_delete",
    ["data_row_id_as_str", "data_row_unique_id", "data_row_global_key"],
)
def test_delete_non_existent_schema_id(
    data_row_for_delete, data_row, mdo, request
):
    res = mdo.bulk_delete(
        [
            DeleteDataRowMetadata(
                data_row_id=request.getfixturevalue(data_row_for_delete),
                fields=[SPLIT_SCHEMA_ID],
            )
        ]
    )
    assert len(res) == 1
    assert res[0].fields == [SPLIT_SCHEMA_ID]
    assert res[0].error == "Schema did not exist"



import json
import random
import threading
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List

import lbox.exceptions
import pytest

from labelbox import Client, DataRow, Dataset
from labelbox.schema.embedding import Embedding


def test_get_embedding_by_id(client: Client, embedding: Embedding):
    e = client.get_embedding_by_id(embedding.id)
    assert e.id == embedding.id

    e = client.get_embedding_by_name(embedding.name)
    assert e.name == embedding.name

    embeddings = client.get_embeddings()
    assert len(embeddings) > 0


def test_get_embedding_by_name_not_found(client: Client):
    with pytest.raises(lbox.exceptions.ResourceNotFoundError):
        client.get_embedding_by_name("does-not-exist")


@pytest.mark.parametrize("data_rows", [10], indirect=True)
def test_import_vectors_from_file(
    data_rows: List[DataRow], embedding: Embedding
):
    vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
    event = threading.Event()

    def callback(_: Dict[str, Any]):
        event.set()

    with NamedTemporaryFile(mode="w+") as fp:
        lines = [
            json.dumps({"id": dr.uid, "vector": vector}) for dr in data_rows
        ]
        fp.writelines(lines)
        fp.flush()
        embedding.import_vectors_from_file(fp.name, callback)

    assert event.wait(10.0)  # seconds


def test_get_imported_vector_count(dataset: Dataset, embedding: Embedding):
    assert embedding.get_imported_vector_count() == 0

    vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
    dataset.create_data_row(
        row_data="foo",
        embeddings=[{"embedding_id": embedding.id, "vector": vector}],
    )

    assert embedding.get_imported_vector_count() == 1



from typing import Generic, TypeVar
import numpy as np
from pydantic_core import core_schema

DType = TypeVar("DType")
DShape = TypeVar("DShape")


class TypedArray(np.ndarray, Generic[DType, DShape]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: type, _model: type
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, val):
        if not isinstance(val, np.ndarray):
            raise TypeError(f"Expected numpy array. Found {type(val)}")
        return val



import logging
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Tuple, Union
import json
import warnings

from labelbox.orm import query
from labelbox.orm.db_object import (
    DbObject,
    Updateable,
    BulkDeletable,
)
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.schema.asset_attachment import AttachmentType
from labelbox.schema.data_row_metadata import DataRowMetadataField  # type: ignore
from labelbox.schema.export_filters import (
    DatarowExportFilters,
    build_filters,
    validate_at_least_one_of_data_row_ids_or_global_keys,
)
from labelbox.schema.export_params import (
    CatalogExportParams,
    validate_catalog_export_params,
)
from labelbox.schema.export_task import ExportTask
from labelbox.schema.task import Task

if TYPE_CHECKING:
    from labelbox import AssetAttachment, Client

logger = logging.getLogger(__name__)


class KeyType(str, Enum):
    ID = "ID"
    """An existing CUID"""
    GKEY = "GKEY"
    """A Global key, could be existing or non-existing"""
    AUTO = "AUTO"
    """The key will be auto-generated. Only usable for creates"""


class DataRow(DbObject, Updateable, BulkDeletable):
    """Internal Labelbox representation of a single piece of data (e.g. image, video, text).

    Attributes:
        external_id (str): User-generated file name or identifier
        global_key (str): User-generated globally unique identifier
        row_data (str): Paths to local files are uploaded to Labelbox's server.
            Otherwise, it's treated as an external URL.
        updated_at (datetime)
        created_at (datetime)
        media_attributes (dict): generated media attributes for the data row
        metadata_fields (list): metadata associated with the data row
        metadata (list): metadata associated with the data row as list of DataRowMetadataField.
            When importing Data Rows with metadata, use `metadata_fields` instead

        dataset (Relationship): `ToOne` relationship to Dataset
        created_by (Relationship): `ToOne` relationship to User
        organization (Relationship): `ToOne` relationship to Organization
        labels (Relationship): `ToMany` relationship to Label
        attachments (Relationship) `ToMany` relationship with AssetAttachment
    """

    external_id = Field.String("external_id")
    global_key = Field.String("global_key")
    row_data = Field.String("row_data")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    media_attributes = Field.Json("media_attributes")
    metadata_fields = Field.List(
        dict,
        graphql_type="DataRowCustomMetadataUpsertInput!",
        name="metadata_fields",
        result_subquery="metadataFields { schemaId name value kind }",
    )
    metadata = Field.List(
        DataRowMetadataField,
        name="metadata",
        graphql_name="customMetadata",
        result_subquery="customMetadata { schemaId value }",
    )

    # Relationships
    dataset = Relationship.ToOne("Dataset")
    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization", False)
    labels = Relationship.ToMany("Label", True)
    attachments = Relationship.ToMany("AssetAttachment", False, "attachments")

    supported_meta_types = supported_attachment_types = set(
        AttachmentType.__members__
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attachments.supports_filtering = False
        self.attachments.supports_sorting = False

    def update(self, **kwargs):
        # Convert row data to string if it is an object
        # All other updates pass through
        primary_fields = ["external_id", "global_key", "row_data"]
        for field in primary_fields:
            data = kwargs.get(field)
            if data == "" or data == {}:
                raise ValueError(f"{field} cannot be empty if it is set")
        if not any(kwargs.get(field) for field in primary_fields):
            raise ValueError(
                f"At least one of these fields needs to be present: {primary_fields}"
            )

        row_data = kwargs.get("row_data")
        if isinstance(row_data, dict):
            kwargs["row_data"] = json.dumps(row_data)
        super().update(**kwargs)

    @staticmethod
    def bulk_delete(data_rows) -> None:
        """Deletes all the given DataRows.

        Args:
            data_rows (list of DataRow): The DataRows to delete.
        """
        BulkDeletable._bulk_delete(data_rows, True)

    def get_winning_label_id(self, project_id: str) -> Optional[str]:
        """Retrieves the winning label ID, i.e. the one that was marked as the
            best for a particular data row, in a project's workflow.

        Args:
            project_id (str): ID of the project containing the data row
        """
        data_row_id_param = "dataRowId"
        project_id_param = "projectId"
        query_str = """query GetWinningLabelIdPyApi($%s: ID!, $%s: ID!) {
            dataRow(where: { id: $%s }) {
                labelingActivity(where: { projectId: $%s }) {
                    selectedLabelId
                }
            }} """ % (
            data_row_id_param,
            project_id_param,
            data_row_id_param,
            project_id_param,
        )

        res = self.client.execute(
            query_str,
            {
                data_row_id_param: self.uid,
                project_id_param: project_id,
            },
        )

        return res["dataRow"]["labelingActivity"]["selectedLabelId"]

    def create_attachment(
        self, attachment_type, attachment_value, attachment_name=None
    ) -> "AssetAttachment":
        """Adds an AssetAttachment to a DataRow.
            Labelers can view these attachments while labeling.

            >>> datarow.create_attachment("TEXT", "This is a text message")

        Args:
            attachment_type (str): Asset attachment type, must be one of:
                VIDEO, IMAGE, TEXT, IMAGE_OVERLAY (AttachmentType)
            attachment_value (str): Asset attachment value.
            attachment_name (str): (Optional) Asset attachment name.
        Returns:
            `AssetAttachment` DB object.
        Raises:
            ValueError: attachment_type must be one of the supported types.
            ValueError: attachment_value must be a non-empty string.
        """
        Entity.AssetAttachment.validate_attachment_json(
            {"type": attachment_type, "value": attachment_value}
        )

        attachment_type_param = "type"
        attachment_value_param = "value"
        attachment_name_param = "name"
        data_row_id_param = "dataRowId"

        query_str = """mutation CreateDataRowAttachmentPyApi(
            $%s: AttachmentType!, $%s: String!, $%s: String, $%s: ID!) {
            createDataRowAttachment(data: {
                type: $%s value: $%s name: $%s dataRowId: $%s}) {%s}} """ % (
            attachment_type_param,
            attachment_value_param,
            attachment_name_param,
            data_row_id_param,
            attachment_type_param,
            attachment_value_param,
            attachment_name_param,
            data_row_id_param,
            query.results_query_part(Entity.AssetAttachment),
        )

        res = self.client.execute(
            query_str,
            {
                attachment_type_param: attachment_type,
                attachment_value_param: attachment_value,
                attachment_name_param: attachment_name,
                data_row_id_param: self.uid,
            },
        )
        return Entity.AssetAttachment(
            self.client, res["createDataRowAttachment"]
        )

    @staticmethod
    def export(
        client: "Client",
        data_rows: Optional[List[Union[str, "DataRow"]]] = None,
        global_keys: Optional[List[str]] = None,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> ExportTask:
        """
        Creates a data rows export task with the given list, params and returns the task.
        Args:
            client (Client): client to use to make the export request
            data_rows (list of DataRow or str): list of data row objects or data row ids to export
            task_name (str): name of remote task
            params (CatalogExportParams): export params

        >>>     dataset = client.get_dataset(DATASET_ID)
        >>>     task = DataRow.export(
        >>>         data_rows=[data_row.uid for data_row in dataset.data_rows.list()],
        >>>             # or a list of DataRow objects: data_rows = data_set.data_rows.list()
        >>>             # or a list of global_keys=["global_key_1", "global_key_2"],
        >>>             # Note that exactly one of: data_rows or global_keys parameters can be passed in at a time
        >>>             # and if data rows ids is present, global keys will be ignored
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """
        task, _ = DataRow._export(
            client, data_rows, global_keys, task_name, params, streamable=True
        )
        return ExportTask(task)

    @staticmethod
    def export_v2(
        client: "Client",
        data_rows: Optional[List[Union[str, "DataRow"]]] = None,
        global_keys: Optional[List[str]] = None,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a data rows export task with the given list, params and returns the task.
        Args:
            client (Client): client to use to make the export request
            data_rows (list of DataRow or str): list of data row objects or data row ids to export
            task_name (str): name of remote task
            params (CatalogExportParams): export params


        >>>     dataset = client.get_dataset(DATASET_ID)
        >>>     task = DataRow.export_v2(
        >>>         data_rows=[data_row.uid for data_row in dataset.data_rows.list()],
        >>>             # or a list of DataRow objects: data_rows = data_set.data_rows.list()
        >>>             # or a list of global_keys=["global_key_1", "global_key_2"],
        >>>             # Note that exactly one of: data_rows or global_keys parameters can be passed in at a time
        >>>             # and if data rows ids is present, global keys will be ignored
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """

        warnings.warn(
            "The method export_v2 for DataRow is deprecated and will be removed in the next major release. Use the export method instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        task, is_streamable = DataRow._export(
            client, data_rows, global_keys, task_name, params
        )
        if is_streamable:
            return ExportTask(task, True)
        return task

    @staticmethod
    def _export(
        client: "Client",
        data_rows: Optional[List[Union[str, "DataRow"]]] = None,
        global_keys: Optional[List[str]] = None,
        task_name: Optional[str] = None,
        params: Optional[CatalogExportParams] = None,
        streamable: bool = False,
    ) -> Tuple[Task, bool]:
        _params = params or CatalogExportParams(
            {
                "attachments": False,
                "embeddings": False,
                "metadata_fields": False,
                "data_row_details": False,
                "project_details": False,
                "performance_details": False,
                "label_details": False,
                "media_type_override": None,
                "model_run_ids": None,
                "project_ids": None,
                "interpolated_frames": False,
                "all_projects": False,
                "all_model_runs": False,
                "predictions": False,
            }
        )

        validate_catalog_export_params(_params)

        mutation_name = "exportDataRowsInCatalog"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInCatalogInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}"
        )

        data_row_ids = []
        if data_rows is not None:
            for dr in data_rows:
                if isinstance(dr, DataRow):
                    data_row_ids.append(dr.uid)
                elif isinstance(dr, str):
                    data_row_ids.append(dr)

        filters = (
            DatarowExportFilters(
                {
                    "data_row_ids": data_row_ids,
                    "global_keys": None,
                }
            )
            if data_row_ids
            else DatarowExportFilters(
                {
                    "data_row_ids": None,
                    "global_keys": global_keys,
                }
            )
        )
        validate_at_least_one_of_data_row_ids_or_global_keys(filters)

        search_query = build_filters(client, filters)
        media_type_override = _params.get("media_type_override", None)

        if task_name is None:
            task_name = f"Export v2: data rows {len(data_row_ids)}"
        query_params = {
            "input": {
                "taskName": task_name,
                "filters": {
                    "searchQuery": {"scope": None, "query": search_query}
                },
                "isStreamableReady": True,
                "params": {
                    "mediaTypeOverride": media_type_override.value
                    if media_type_override is not None
                    else None,
                    "includeAttachments": _params.get("attachments", False),
                    "includeEmbeddings": _params.get("embeddings", False),
                    "includeMetadata": _params.get("metadata_fields", False),
                    "includeDataRowDetails": _params.get(
                        "data_row_details", False
                    ),
                    "includeProjectDetails": _params.get(
                        "project_details", False
                    ),
                    "includePerformanceDetails": _params.get(
                        "performance_details", False
                    ),
                    "includeLabelDetails": _params.get("label_details", False),
                    "includeInterpolatedFrames": _params.get(
                        "interpolated_frames", False
                    ),
                    "projectIds": _params.get("project_ids", None),
                    "modelRunIds": _params.get("model_run_ids", None),
                    "allProjects": _params.get("all_projects", False),
                    "allModelRuns": _params.get("all_model_runs", False),
                },
                "streamable": streamable,
            }
        }

        res = client.execute(
            create_task_query_str, query_params, error_log_key="errors"
        )
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(client, task_id), is_streamable



import json
from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.mixins import CustomMetric
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import ObjectAnnotation, Point, Line, Label


def test_polyline_import_with_confidence():
    with open(
        "tests/data/assets/ndjson/polyline_without_confidence_import.json", "r"
    ) as file:
        data = json.load(file)
    labels = [
        Label(
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    name="some-line",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=Line(
                        points=[
                            Point(x=2534.353, y=249.471),
                            Point(x=2429.492, y=182.092),
                            Point(x=2294.322, y=221.962),
                        ],
                    ),
                )
            ],
        )
    ]
    res = list(NDJsonConverter.serialize(labels))
    assert res == data


def test_polyline_import_without_confidence():
    with open("tests/data/assets/ndjson/polyline_import.json", "r") as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="cl6xnv9h61fv0085yhtoq06ht",
            ),
            annotations=[
                ObjectAnnotation(
                    custom_metrics=[
                        CustomMetric(name="customMetric1", value=0.5),
                        CustomMetric(name="customMetric2", value=0.3),
                    ],
                    confidence=0.58,
                    name="some-line",
                    feature_schema_id="cl6xnuwt95lqq07330tbb3mfd",
                    extra={
                        "uuid": "5ad9c52f-058d-49c8-a749-3f20b84f8cd4",
                    },
                    value=Line(
                        points=[
                            Point(x=2534.353, y=249.471),
                            Point(x=2429.492, y=182.092),
                            Point(x=2294.322, y=221.962),
                        ],
                    ),
                )
            ],
        )
    ]

    res = list(NDJsonConverter.serialize(labels))
    assert res == data



from typing import Dict, Optional, TypedDict

from pydantic import BaseModel, model_validator

from labelbox.schema.conflict_resolution_strategy import (
    ConflictResolutionStrategy,
)


class SendToAnnotateFromCatalogParams(BaseModel):
    """
    Extra parameters for sending data rows to a project through catalog. At least one of source_model_run_id or
    source_project_id must be provided.

    :param source_model_run_id: Optional[str] - The model run to use for predictions. Defaults to None.
    :param predictions_ontology_mapping: Optional[Dict[str, str]] - A mapping of feature schema ids to feature schema
        ids. Defaults to an empty dictionary.
    :param source_project_id: Optional[str] - The project to use for predictions. Defaults to None.
    :param annotations_ontology_mapping: Optional[Dict[str, str]] - A mapping of feature schema ids to feature schema
        ids. Defaults to an empty dictionary.
    :param exclude_data_rows_in_project: Optional[bool] - Exclude data rows that are already in the project. Defaults
        to False.
    :param override_existing_annotations_rule: Optional[ConflictResolutionStrategy] - The strategy defining how to
        handle conflicts in classifications between the data rows that already exist in the project and incoming
        predictions from the source model run or annotations from the source project. Defaults to
        ConflictResolutionStrategy.KEEP_EXISTING.
    :param batch_priority: Optional[int] - The priority of the batch. Defaults to 5.
    """

    source_model_run_id: Optional[str] = None
    source_project_id: Optional[str] = None
    predictions_ontology_mapping: Optional[Dict[str, str]] = {}
    annotations_ontology_mapping: Optional[Dict[str, str]] = {}
    exclude_data_rows_in_project: Optional[bool] = False
    override_existing_annotations_rule: Optional[ConflictResolutionStrategy] = (
        ConflictResolutionStrategy.KeepExisting
    )
    batch_priority: Optional[int] = 5

    @model_validator(mode="after")
    def check_project_id_or_model_run_id(self):
        if not self.source_model_run_id and not self.source_project_id:
            raise ValueError(
                "Either source_project_id or source_model_id are required"
            )
        if self.source_model_run_id and self.source_project_id:
            raise ValueError(
                "Provide only a source_project_id or source_model_id not both"
            )
        return self


class SendToAnnotateFromModelParams(TypedDict):
    """
    Extra parameters for sending data rows to a project through a model run.

    :param predictions_ontology_mapping: Dict[str, str] - A mapping of feature schema ids to feature schema ids.
        Defaults to an empty dictionary.
    :param exclude_data_rows_in_project: Optional[bool] - Exclude data rows that are already in the project. Defaults
        to False.
    :param override_existing_annotations_rule: Optional[ConflictResolutionStrategy] - The strategy defining how to
        handle conflicts in classifications between the data rows that already exist in the project and incoming
        predictions from the source model run. Defaults to ConflictResolutionStrategy.KEEP_EXISTING.
    :param batch_priority: Optional[int] - The priority of the batch. Defaults to 5.
    """

    predictions_ontology_mapping: Dict[str, str]
    exclude_data_rows_in_project: Optional[bool]
    override_existing_annotations_rule: Optional[ConflictResolutionStrategy]
    batch_priority: Optional[int]


def build_annotations_input(
    project_ontology_mapping: Optional[Dict[str, str]], source_project_id: str
):
    return {
        "projectId": source_project_id,
        "featureSchemaIdsMapping": project_ontology_mapping
        if project_ontology_mapping
        else {},
    }


def build_destination_task_queue_input(task_queue_id: str):
    destination_task_queue = (
        {"type": "id", "value": task_queue_id}
        if task_queue_id
        else {"type": "done"}
    )
    return destination_task_queue


def build_predictions_input(
    model_run_ontology_mapping: Optional[Dict[str, str]],
    source_model_run_id: str,
):
    return {
        "featureSchemaIdsMapping": model_run_ontology_mapping
        if model_run_ontology_mapping
        else {},
        "modelRunId": source_model_run_id,
        "minConfidence": 0,
        "maxConfidence": 1,
    }



from .geometry import Line
from .geometry import Point
from .geometry import Mask
from .geometry import Polygon
from .geometry import Rectangle
from .geometry import Geometry
from .geometry import DocumentRectangle
from .geometry import RectangleUnit

from .annotation import ClassificationAnnotation
from .annotation import ObjectAnnotation

from .relationship import RelationshipAnnotation
from .relationship import Relationship

from .video import VideoClassificationAnnotation
from .video import VideoObjectAnnotation
from .video import MaskFrame
from .video import MaskInstance
from .video import VideoMaskAnnotation

from .ner import ConversationEntity
from .ner import DocumentEntity
from .ner import DocumentTextSelection
from .ner import TextEntity

from .classification import Checklist
from .classification import ClassificationAnswer
from .classification import Radio
from .classification import Text

from .data import GenericDataRowData
from .data import MaskData

from .label import Label
from .collection import LabelGenerator

from .metrics import ScalarMetric
from .metrics import ScalarMetricAggregation
from .metrics import ConfusionMatrixMetric
from .metrics import ConfusionMatrixAggregation
from .metrics import ScalarMetricValue
from .metrics import ConfusionMatrixMetricValue

from .data.tiled_image import EPSG
from .data.tiled_image import EPSGTransformer
from .data.tiled_image import TiledBounds
from .data.tiled_image import TiledImageData
from .data.tiled_image import TileLayer

from .llm_prompt_response.prompt import PromptText
from .llm_prompt_response.prompt import PromptClassificationAnnotation

from .mmc import (
    MessageInfo,
    OrderedMessageInfo,
    MessageSingleSelectionTask,
    MessageMultiSelectionTask,
    MessageRankingTask,
    MessageEvaluationTaskAnnotation,
)



from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Dict, List, Type, Union

from lbox.exceptions import InvalidAttributeError

import labelbox
from labelbox import utils
from labelbox.orm.comparison import Comparison

""" Defines Field, Relationship and Entity. These classes are building
blocks for defining the Labelbox schema, DB object operations and
queries. """


class Field:
    """Represents a field in a database table. A Field has a name, a type
    (corresponds to server-side GraphQL type) and a server-side name. The
    server-side name is most often just a camelCase version of the client-side
    snake_case name.

    Supports comparison operators which return a `labelbox.comparison.Comparison`
    object. For example:
        >>> class Project:
        >>>     name = Field.String("name")
        >>>
        >>> comparison = Project.name == "MyProject"

    These `Comparison` objects can then be used for filtering:
        >>> project = client.get_projects(comparison)

    Also exposes the ordering property used for sorting:
        >>> labels = project.labels(order_by=Label.label.asc)

    Attributes:
        field_type (Field.Type): The type of the field.
        name (str): name that the attribute has in client-side Python objects
        graphql_name (str): name that the attribute has in queries (and in
            server-side database definition).
        result_subquery (str): graphql query result payload for a field.
    """

    class Type(Enum):
        Int = auto()
        Float = auto()
        String = auto()
        Boolean = auto()
        ID = auto()
        DateTime = auto()
        Json = auto()

    class EnumType:
        def __init__(self, enum_cls: type):
            self.enum_cls = enum_cls

        @property
        def name(self):
            return self.enum_cls.__name__

    class ListType:
        """Represents Field that is a list of some object.
        Args:
            list_cls (type): Type of object that list is made of.
            graphql_type (str): Inner object's graphql type.
                By default, the list_cls's name is used as the graphql type.
        """

        def __init__(self, list_cls: type, graphql_type=None):
            self.list_cls = list_cls
            if graphql_type is None:
                self.graphql_type = self.list_cls.__name__
            else:
                self.graphql_type = graphql_type

        @property
        def name(self):
            return f"[{self.graphql_type}]"

    class Order(Enum):
        """Type of sort ordering."""

        Asc = auto()
        Desc = auto()

    @staticmethod
    def Int(*args):
        return Field(Field.Type.Int, *args)

    @staticmethod
    def Float(*args):
        return Field(Field.Type.Float, *args)

    @staticmethod
    def String(*args):
        return Field(Field.Type.String, *args)

    @staticmethod
    def Boolean(*args):
        return Field(Field.Type.Boolean, *args)

    @staticmethod
    def ID(*args):
        return Field(Field.Type.ID, *args)

    @staticmethod
    def DateTime(*args):
        return Field(Field.Type.DateTime, *args)

    @staticmethod
    def Enum(enum_cls: type, *args):
        return Field(Field.EnumType(enum_cls), *args)

    @staticmethod
    def Json(*args):
        return Field(Field.Type.Json, *args)

    @staticmethod
    def List(list_cls: type, graphql_type=None, **kwargs):
        return Field(Field.ListType(list_cls, graphql_type), **kwargs)

    def __init__(
        self,
        field_type: Union[Type, EnumType, ListType],
        name,
        graphql_name=None,
        result_subquery=None,
    ):
        """Field init.
        Args:
            field_type (Field.Type): The type of the field.
            name (str): client-side Python attribute name of a database
                object.
            graphql_name (str): query and server-side name of a database object.
                If None, it is constructed from the client-side name by converting
                snake_case (Python convention) into camelCase (GraphQL convention).
            result_subquery (str): graphql query result payload for a field.
        """
        self.field_type = field_type
        self.name = name
        if graphql_name is None:
            graphql_name = utils.camel_case(name)
        self.graphql_name = graphql_name
        self.result_subquery = result_subquery

    @property
    def asc(self):
        """Property that resolves to tuple (Field, Field.Order).
        Used for easy definition of sort ordering:
            >>> projects_ordered = client.get_projects(order_by=Project.name.asc)
        """
        return (self, Field.Order.Asc)

    @property
    def desc(self):
        """Property that resolves to tuple (Field, Field.Order).
        Used for easy definition of sort ordering:
            >>> projects_ordered = client.get_projects(order_by=Project.name.desc)
        """
        return (self, Field.Order.Desc)

    def __eq__(self, other):
        """Equality of Fields has two meanings. If comparing to a Field object,
        then a boolean indicator if the fields are identical is returned. If
        comparing to any other type, a Comparison object is created.
        """
        if isinstance(other, Field):
            return self is other

        return Comparison.Op.EQ(self, other)

    def __ne__(self, other):
        """Equality of Fields has two meanings. If comparing to a Field object,
        then a boolean indicator if the fields are identical is returned. If
        comparing to any other type, a Comparison object is created.
        """
        if isinstance(other, Field):
            return self is not other

        return Comparison.Op.NE(self, other)

    def __hash__(self):
        # Hash is implemeted as ID, because for each DB field exactly one
        # Field object should exist in the Python API.
        return id(self)

    def __lt__(self, other):
        return Comparison.Op.LT(self, other)

    def __gt__(self, other):
        return Comparison.Op.GT(self, other)

    def __le__(self, other):
        return Comparison.Op.LE(self, other)

    def __ge__(self, other):
        return Comparison.Op.GE(self, other)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Field: %r>" % self.name


class Relationship:
    """Represents a relationship in a database table.

    Attributes:
        relationship_type (Relationship.Type): Indicator if to-one or to-many
        destination_type_name (str): Name of the Entity subtype that's on
            the other side of the relationship. str is used instead of the
            type object itself because that type might not be declared at
            the point of a `Relationship` object initialization.
        filter_deleted (bool): Indicator if the a `deleted=false` filtering
            clause should be added to the query when fetching relationship
            objects.
        name (str): Name of the relationship in the snake_case format.
        graphql_name (str): Name of the relationships server-side. Most often
            (not always) just a camelCase version of `name`.
        cache (bool) : Whether or not to cache the relationship values.
            Useful for objects that aren't directly queryable from the api (relationship query builder won't work)
            Also useful for expensive ToOne relationships
        deprecation_warning (string) optional message to display when RelationshipManager is called

    """

    @dataclass
    class Config:
        disconnect_supported: bool = True

    class Type(Enum):
        ToOne = auto()
        ToMany = auto()

    @staticmethod
    def ToOne(*args, **kwargs):
        return Relationship(Relationship.Type.ToOne, *args, **kwargs)

    @staticmethod
    def ToMany(*args, **kwargs):
        return Relationship(Relationship.Type.ToMany, *args, **kwargs)

    def __init__(
        self,
        relationship_type,
        destination_type_name,
        filter_deleted=True,
        name=None,
        graphql_name=None,
        cache=False,
        deprecation_warning=None,
        config=Config(),
    ):
        self.relationship_type = relationship_type
        self.destination_type_name = destination_type_name
        self.filter_deleted = filter_deleted
        self.cache = cache
        self.deprecation_warning = deprecation_warning
        self.config = config

        if name is None:
            name = utils.snake_case(destination_type_name) + (
                "s" if relationship_type == Relationship.Type.ToMany else ""
            )
        self.name = name

        if graphql_name is None:
            graphql_name = utils.camel_case(name)
        self.graphql_name = graphql_name

    @property
    def destination_type(self):
        return getattr(Entity, self.destination_type_name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Relationship: %r>" % self.name


class EntityMeta(type):
    """Entity metaclass. Registers Entity subclasses as attributes
    of the Entity class object so they can be referenced for example like:
        Entity.Project.
    """

    # Maps Entity name to Relationships for all currently defined Entities
    relationship_mappings: Dict[str, List[Relationship]] = {}

    def __setattr__(self, key: Any, value: Any):
        super().__setattr__(key, value)

    def __init__(cls, clsname, superclasses, attributedict):
        super().__init__(clsname, superclasses, attributedict)
        cls.validate_cached_relationships()
        if clsname != "Entity":
            setattr(Entity, clsname, cls)
            EntityMeta.relationship_mappings[utils.snake_case(cls.__name__)] = (
                cls.relationships()
            )

    @staticmethod
    def raise_for_nested_cache(first: str, middle: str, last: List[str]):
        raise TypeError(
            "Cannot cache a relationship to an Entity with its own cached relationship(s). "
            f"`{first}` caches `{middle}` which caches `{last}`"
        )

    @staticmethod
    def cached_entities(entity_name: str):
        """
        Return all cached entites for a given Entity name
        """
        cached_entities = EntityMeta.relationship_mappings.get(entity_name, [])
        return {
            entity.name: entity for entity in cached_entities if entity.cache
        }

    def validate_cached_relationships(cls):
        """
        Graphql doesn't allow for infinite nesting in queries.
        This function checks that cached relationships result in valid queries.
            * It does this by making sure that a cached relationship do not
              reference any entity with its own cached relationships.

        This check is performed by looking to see if this entity caches
        any entities that have their own cached fields. If this entity
        that we are checking has any cached fields then we also check
        all currently defined entities to see if they cache this entity.

        A two way check is necessary because checks are performed as classes are being defined.
        As opposed to after all objects have been created.
        """
        # All cached relationships
        cached_rels = [r for r in cls.relationships() if r.cache]

        # Check if any cached entities have their own cached fields
        for rel in cached_rels:
            nested = cls.cached_entities(rel.name)
            if nested:
                cls.raise_for_nested_cache(
                    utils.snake_case(cls.__name__),
                    rel.name,
                    list(nested.keys()),
                )

        # If the current Entity (cls) has any cached relationships (cached_rels)
        #  then no other defined Entity (entities in EntityMeta.relationship_mappings) can cache this Entity.
        if cached_rels:
            # For all currently defined Entities
            for entity_name in EntityMeta.relationship_mappings:
                # Get all cached ToOne relationships
                rels = cls.cached_entities(entity_name)
                # Check if the current Entity (cls) is referenced by the Entity with `entity_name`
                rel = rels.get(utils.snake_case(cls.__name__))
                # If rel exists and is cached then raise an exception
                # This means `entity_name` caches `cls` which cached items in `cached_rels`
                if rel and rel.cache:
                    cls.raise_for_nested_cache(
                        utils.snake_case(entity_name),
                        utils.snake_case(cls.__name__),
                        [entity.name for entity in cached_rels],
                    )


class Entity(metaclass=EntityMeta):
    """An entity that contains fields and relationships. Base class
    for DbObject (which is base class for concrete schema classes)."""

    # Every Entity has an "id" and a "deleted" field
    # Name the "id" field "uid" in Python to avoid conflict with keyword.
    uid = Field.ID("uid", "id")

    # Some Labelbox objects have a "deleted" attribute for soft deletions.
    # It's declared in Entity so it can be filtered out in class methods
    # suchs as `fields()`.
    deleted = Field.Boolean("deleted")

    if TYPE_CHECKING:
        DataRow: Type[labelbox.DataRow]
        Webhook: Type[labelbox.Webhook]
        Task: Type[labelbox.Task]
        AssetAttachment: Type[labelbox.AssetAttachment]
        ModelRun: Type[labelbox.ModelRun]
        ModelConfig: Type[labelbox.ModelConfig]
        Review: Type[labelbox.Review]
        User: Type[labelbox.User]
        LabelingFrontend: Type[labelbox.LabelingFrontend]
        Benchmark: Type[labelbox.Benchmark]
        IAMIntegration: Type[labelbox.IAMIntegration]
        LabelingFrontendOptions: Type[labelbox.LabelingFrontendOptions]
        Label: Type[labelbox.Label]
        MEAPredictionImport: Type[labelbox.MEAPredictionImport]
        MALPredictionImport: Type[labelbox.MALPredictionImport]
        Invite: Type[labelbox.Invite]
        InviteLimit: Type[labelbox.InviteLimit]
        ProjectRole: Type[labelbox.ProjectRole]
        ProjectModelConfig: Type[labelbox.ProjectModelConfig]
        Project: Type[labelbox.Project]
        Batch: Type[labelbox.Batch]
        CatalogSlice: Type[labelbox.CatalogSlice]
        ModelSlice: Type[labelbox.ModelSlice]
        TaskQueue: Type[labelbox.TaskQueue]

    @classmethod
    def _attributes_of_type(cls, attr_type):
        """Yields all the attributes in `cls` of the given `attr_type`."""
        for attr_name in dir(cls):
            attr = getattr(cls, attr_name)
            if isinstance(attr, attr_type):
                yield attr

    @classmethod
    def fields(cls):
        """Returns a generator that yields all the Fields declared in a
        concrete subclass.
        """
        for attr in cls._attributes_of_type(Field):
            if attr != Entity.deleted:
                yield attr

    @classmethod
    def relationships(cls):
        """Returns a generator that yields all the Relationships declared in
        a concrete subclass.
        """
        return cls._attributes_of_type(Relationship)

    @classmethod
    def field(cls, field_name):
        """Returns a Field object for the given name.
        Args:
            field_name (str): Field name, Python (snake-case) convention.
        Return:
            Field object
        Raises:
            InvalidAttributeError: in case this DB object type does not
            contain a field with the given name.
        """
        field_obj = getattr(cls, field_name, None)
        if not isinstance(field_obj, Field):
            raise InvalidAttributeError(cls, field_name)
        return field_obj

    @classmethod
    def attribute(cls, attribute_name):
        """Returns a Field or a Relationship object for the given name.
        Args:
            attribute_name (str): Field or Relationship name, Python
                (snake-case) convention.
        Return:
            Field or Relationship object
        Raises:
            InvalidAttributeError: in case this DB object type does not
                contain an attribute with the given name.
        """
        attribute_object = getattr(cls, attribute_name, None)
        if not isinstance(attribute_object, (Field, Relationship)):
            raise InvalidAttributeError(cls, attribute_name)
        return attribute_object

    @classmethod
    def type_name(cls):
        """Returns this DB object type name in TitleCase. For example:
        Project, DataRow, ...
        """
        return cls.__name__.split(".")[-1]



import pytest
import cv2

from labelbox.data.annotation_types.geometry import Point, Line
from pydantic import ValidationError


def test_line():
    with pytest.raises(ValidationError):
        line = Line()

    with pytest.raises(ValidationError):
        line = Line(points=[[0, 1], [2, 3]])

    points = [[0, 1], [0, 2], [2, 2]]
    expected = {"coordinates": [points], "type": "MultiLineString"}
    line = Line(points=[Point(x=x, y=y) for x, y in points])
    assert line.geometry == expected
    expected["coordinates"] = tuple([tuple([tuple(x) for x in points])])
    assert line.shapely.__geo_interface__ == expected

    raster = line.draw(height=32, width=32, thickness=1)
    assert (cv2.imread("tests/data/assets/line.png") == raster).all()



from abc import ABC, abstractmethod

from typing import List, Tuple, Optional

from labelbox.schema.identifiable import UniqueId, GlobalKey
from labelbox.schema.data_row import DataRow
from pydantic import BaseModel


class DataRowItemBase(ABC, BaseModel):
    """
    Base class for creating payloads for upsert operations.
    """

    id: dict
    payload: dict

    @abstractmethod
    def is_empty(self) -> bool: ...

    @classmethod
    def build(
        cls,
        dataset_id: str,
        items: List[dict],
        key_types: Optional[Tuple[type, ...]] = (),
    ) -> List["DataRowItemBase"]:
        upload_items = []

        for item in items:
            # enforce current dataset's id for all specs
            item["dataset_id"] = dataset_id
            key = item.pop("key", None)
            if not key:
                key = {"type": "AUTO", "value": ""}
            elif isinstance(key, key_types):  # type: ignore
                key = {"type": key.id_type.value, "value": key.key}
            else:
                if not key_types:
                    raise ValueError(
                        f"Can not have a key for this item, got: {key}"
                    )
                raise ValueError(
                    f"Key must be an instance of {', '.join([t.__name__ for t in key_types])}, got: {type(item['key']).__name__}"
                )
            item = {
                k: v for k, v in item.items() if v is not None
            }  # remove None values
            upload_items.append(cls(payload=item, id=key))
        return upload_items


class DataRowUpsertItem(DataRowItemBase):
    def is_empty(self) -> bool:
        """
        The payload is considered empty if it's actually empty or the only key is `dataset_id`.
        :return: bool
        """
        return (
            not self.payload
            or len(self.payload.keys()) == 1
            and "dataset_id" in self.payload
        )

    @classmethod
    def build(
        cls,
        dataset_id: str,
        items: List[dict],
        key_types: Optional[Tuple[type, ...]] = (UniqueId, GlobalKey),
    ) -> List["DataRowItemBase"]:
        return super().build(dataset_id, items, (UniqueId, GlobalKey))


class DataRowCreateItem(DataRowItemBase):
    def is_empty(self) -> bool:
        """
        The payload is considered empty if it's actually empty or row_data is empty
            or the only key is `dataset_id`.
        :return: bool
        """
        row_data = self.payload.get("row_data", None) or self.payload.get(
            DataRow.row_data, None
        )

        return not self._is_legacy_conversational_data() and (
            not self.payload
            or len(self.payload.keys()) == 1
            and "dataset_id" in self.payload
            or row_data is None
            or len(row_data) == 0
        )

    def _is_legacy_conversational_data(self) -> bool:
        return (
            "conversationalData" in self.payload.keys()
            or "conversational_data" in self.payload.keys()
        )

    @classmethod
    def build(
        cls,
        dataset_id: str,
        items: List[dict],
        key_types: Optional[Tuple[type, ...]] = (),
    ) -> List["DataRowItemBase"]:
        return super().build(dataset_id, items, ())



from labelbox.utils import _CamelCaseMixin


class ConsensusSettings(_CamelCaseMixin):
    """Container for holding consensus quality settings

    >>> ConsensusSettings(
    >>>    number_of_labels = 2,
    >>>    coverage_percentage = 0.2
    >>>  )

    Args:
        number_of_labels: Number of labels for consensus
        coverage_percentage: Percentage of data rows to be labeled more than once
    """

    number_of_labels: int
    coverage_percentage: float



from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship


class LabelingFrontend(DbObject):
    """Private db object representing a projects label editor"""

    name = Field.String("name")
    description = Field.String("description")
    iframe_url_path = Field.String("iframe_url_path")


class LabelingFrontendOptions(DbObject):
    """Label interface options.

    Attributes:
        customization_options (str)

        project (Relationship): `ToOne` relationship to Project
        labeling_frontend (Relationship): `ToOne` relationship to LabelingFrontend
        organization (Relationship): `ToOne` relationship to Organization
    """

    customization_options = Field.String("customization_options")

    project = Relationship.ToOne("Project")
    labeling_frontend = Relationship.ToOne("LabelingFrontend")
    organization = Relationship.ToOne("Organization")



from labelbox import UniqueIds, OntologyBuilder
from labelbox.schema.conflict_resolution_strategy import (
    ConflictResolutionStrategy,
)


def test_send_to_annotate_from_model(
    client,
    configured_project,
    model_run_predictions,
    model_run_with_data_rows,
    project,
):
    model_run = model_run_with_data_rows
    data_row_ids = list(
        set([p["dataRow"]["id"] for p in model_run_predictions])
    )
    assert len(data_row_ids) > 0

    destination_project = project
    model = client.get_model(model_run.model_id)
    ontology = client.get_ontology(model.ontology_id)
    destination_project.connect_ontology(ontology)

    queues = destination_project.task_queues()
    initial_review_task = next(
        q for q in queues if q.name == "Initial review task"
    )

    # build an ontology mapping using the top level tools and classifications
    source_ontology_builder = OntologyBuilder.from_project(configured_project)
    feature_schema_ids = list(
        tool.feature_schema_id for tool in source_ontology_builder.tools
    )
    # create a dictionary of feature schema id to itself
    ontology_mapping = dict(zip(feature_schema_ids, feature_schema_ids))

    classification_feature_schema_ids = list(
        classification.feature_schema_id
        for classification in source_ontology_builder.classifications
    )
    # create a dictionary of feature schema id to itself
    classification_ontology_mapping = dict(
        zip(
            classification_feature_schema_ids, classification_feature_schema_ids
        )
    )

    # combine the two ontology mappings
    ontology_mapping.update(classification_ontology_mapping)

    task = model_run.send_to_annotate_from_model(
        destination_project_id=destination_project.uid,
        batch_name="batch",
        data_rows=UniqueIds(data_row_ids),
        task_queue_id=initial_review_task.uid,
        params={
            "predictions_ontology_mapping": ontology_mapping,
            "override_existing_annotations_rule": ConflictResolutionStrategy.OverrideWithPredictions,
        },
    )

    task.wait_till_done()

    # Check that the data row was sent to the new project
    destination_batches = list(destination_project.batches())
    assert len(destination_batches) == 1

    export_task = destination_project.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()

    destination_data_rows = [dr.json["data_row"]["id"] for dr in stream]

    assert len(destination_data_rows) == len(data_row_ids)
    assert all([dr in data_row_ids for dr in destination_data_rows])

    # Since data rows were added to a review queue, predictions should be imported into the project as labels
    destination_project_labels = list(destination_project.labels())
    assert len(destination_project_labels) == len(data_row_ids)



from abc import ABC
from io import BytesIO
from typing import Callable, Literal, Optional, Union

import numpy as np
import requests
from google.api_core import retry
from lbox.exceptions import InternalServerError
from PIL import Image
from pydantic import BaseModel, ConfigDict, model_validator
from requests.exceptions import ConnectTimeout

from ..types import TypedArray


class RasterData(BaseModel, ABC):
    """Represents an image or segmentation mask."""

    im_bytes: Optional[bytes] = None
    file_path: Optional[str] = None
    url: Optional[str] = None
    uid: Optional[str] = None
    global_key: Optional[str] = None
    arr: Optional[TypedArray[Literal["uint8"]]] = None

    model_config = ConfigDict(extra="forbid")

    @classmethod
    def from_2D_arr(
        cls,
        arr: Union[TypedArray[Literal["uint8"]], TypedArray[Literal["int"]]],
        **kwargs,
    ) -> "RasterData":
        """Construct from a 2D numpy array

        Args:
            arr: uint8 compatible numpy array

        Returns:
            RasterData
        """

        if len(arr.shape) != 2:
            raise ValueError(
                f"Found array with shape {arr.shape}. Expected two dimensions [H, W]"
            )

        if not np.issubdtype(arr.dtype, np.integer):
            raise ValueError("Array must be an integer subtype")

        if np.can_cast(arr, np.uint8):
            arr = arr.astype(np.uint8)
        else:
            raise ValueError(
                "Could not cast array to uint8, check that values are between 0 and 255"
            )

        arr = np.stack((arr,) * 3, axis=-1)
        return cls(arr=arr, **kwargs)

    def bytes_to_np(self, image_bytes: bytes) -> np.ndarray:
        """
        Converts image bytes to a numpy array
        Args:
            image_bytes (bytes): PNG encoded image
        Returns:
            numpy array representing the image
        """
        arr = np.array(Image.open(BytesIO(image_bytes)))
        if len(arr.shape) == 2:
            arr = np.stack((arr,) * 3, axis=-1)
        return arr[:, :, :3]

    def np_to_bytes(self, arr: np.ndarray) -> bytes:
        """
        Converts a numpy array to bytes
        Args:
            arr (np.array): numpy array representing the image
        Returns:
            png encoded bytes
        """
        if len(arr.shape) != 3:
            raise ValueError(
                "unsupported image format. Must be 3D ([H,W,C])."
                f"Use {self.__class__.__name__}.from_2D_arr to construct from 2D"
            )
        if arr.dtype != np.uint8:
            raise TypeError(f"image data type must be uint8. Found {arr.dtype}")

        im_bytes = BytesIO()
        Image.fromarray(arr).save(im_bytes, format="PNG")
        return im_bytes.getvalue()

    @property
    def value(self) -> np.ndarray:
        """
        Property that unifies the data access pattern for all references to the raster.

        Returns:
            numpy representation of the raster
        """
        if self.arr is not None:
            return self.arr
        if self.im_bytes is not None:
            return self.bytes_to_np(self.im_bytes)
        elif self.file_path is not None:
            with open(self.file_path, "rb") as img:
                im_bytes = img.read()
            self.im_bytes = im_bytes
            arr = self.bytes_to_np(im_bytes)
            return arr
        elif self.url is not None:
            im_bytes = self.fetch_remote()
            self.im_bytes = im_bytes
            return self.bytes_to_np(im_bytes)
        else:
            raise ValueError("Must set either url, file_path or im_bytes")

    def set_fetch_fn(self, fn):
        object.__setattr__(self, "fetch_remote", lambda: fn(self))

    @retry.Retry(
        deadline=15.0,
        predicate=retry.if_exception_type(ConnectTimeout, InternalServerError),
    )
    def fetch_remote(self) -> bytes:
        """
        Method for accessing url.

        If url is not publicly accessible or requires another access pattern
        simply override this function
        """
        response = requests.get(self.url)
        if response.status_code in [500, 502, 503, 504]:
            raise InternalServerError(response.text)
        response.raise_for_status()
        return response.content

    @retry.Retry(deadline=30.0)
    def create_url(self, signer: Callable[[bytes], str]) -> str:
        """
        Utility for creating a url from any of the other image representations.

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            url for the raster data
        """
        if self.url is not None:
            return self.url
        elif self.im_bytes is not None:
            self.url = signer(self.im_bytes)
        elif self.file_path is not None:
            with open(self.file_path, "rb") as file:
                self.url = signer(file.read())
        elif self.arr is not None:
            self.url = signer(self.np_to_bytes(self.arr))
        else:
            raise ValueError(
                "One of url, im_bytes, file_path, arr must not be None."
            )
        return self.url

    @model_validator(mode="after")
    def validate_args(self, values):
        file_path = self.file_path
        im_bytes = self.im_bytes
        url = self.url
        arr = self.arr
        uid = self.uid
        global_key = self.global_key
        if (
            uid == file_path == im_bytes == url == global_key is None
            and arr is None
        ):
            raise ValueError(
                "One of `file_path`, `im_bytes`, `url`, `uid`, `global_key` or `arr` required."
            )
        if arr is not None:
            if arr.dtype != np.uint8:
                raise TypeError(
                    "Numpy array representing segmentation mask must be np.uint8"
                )
            elif len(arr.shape) != 3:
                raise ValueError(
                    "unsupported image format. Must be 3D ([H,W,C])."
                    f"Use {self.__name__}.from_2D_arr to construct from 2D"
                )
        return self

    def __repr__(self) -> str:
        def symbol_or_none(data):
            return "..." if data is not None else None

        return (
            f"{self.__class__.__name__}(im_bytes={symbol_or_none(self.im_bytes)},"
            f"file_path={self.file_path},"
            f"url={self.url},"
            f"arr={symbol_or_none(self.arr)})"
        )


class MaskData(RasterData):
    """Used to represent a segmentation Mask

    All segments within a mask must be mutually exclusive. At a
    single cell, only one class can be present. All Mask data is
    converted to a [H,W,3] image. Classes are

    >>> # 3x3 mask with two classes and back ground
    >>> MaskData.from_2D_arr([
    >>>    [0, 0, 0],
    >>>    [1, 1, 1],
    >>>    [2, 2, 2],
    >>>])

    Args:
        im_bytes: Optional[bytes] = None
        file_path: Optional[str] = None
        url: Optional[str] = None
        arr: Optional[TypedArray[Literal['uint8']]] = None
    """



from labelbox import utils

SNAKE = "this_is_a_string"
TITLE = "ThisIsAString"
CAMEL = "thisIsAString"
MIXED = "this_Is_AString"


def test_camel():
    assert utils.camel_case(SNAKE) == CAMEL
    assert utils.camel_case(TITLE) == CAMEL
    assert utils.camel_case(CAMEL) == CAMEL
    assert utils.camel_case(MIXED) == CAMEL


def test_snake():
    assert utils.snake_case(SNAKE) == SNAKE
    assert utils.snake_case(TITLE) == SNAKE
    assert utils.snake_case(CAMEL) == SNAKE
    assert utils.snake_case(MIXED) == SNAKE


def test_title():
    assert utils.title_case(SNAKE) == TITLE
    assert utils.title_case(TITLE) == TITLE
    assert utils.title_case(CAMEL) == TITLE
    assert utils.title_case(MIXED) == TITLE



import uuid
from datetime import datetime, timezone

import pytest
from lbox.exceptions import MalformedQueryException

from labelbox import Dataset
from labelbox.schema.data_row_metadata import (
    DataRowMetadata,
    DataRowMetadataField,
    DataRowMetadataKind,
    DataRowMetadataOntology,
    _parse_metadata_schema,
)
from labelbox.schema.identifiables import GlobalKeys, UniqueIds

INVALID_SCHEMA_ID = "1" * 25
FAKE_SCHEMA_ID = "0" * 25
FAKE_DATAROW_ID = "D" * 25
SPLIT_SCHEMA_ID = "cko8sbczn0002h2dkdaxb5kal"
TRAIN_SPLIT_ID = "cko8sbscr0003h2dk04w86hof"
TEST_SPLIT_ID = "cko8scbz70005h2dkastwhgqt"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"
CAPTURE_DT_SCHEMA_ID = "cko8sdzv70006h2dk8jg64zvb"
CUSTOM_TEXT_SCHEMA_NAME = "custom_text"

FAKE_NUMBER_FIELD = {
    "id": FAKE_SCHEMA_ID,
    "name": "number",
    "kind": "CustomMetadataNumber",
    "reserved": False,
}


@pytest.fixture
def mdo(client):
    mdo = client.get_data_row_metadata_ontology()
    try:
        mdo.create_schema(CUSTOM_TEXT_SCHEMA_NAME, DataRowMetadataKind.string)
    except MalformedQueryException:
        # Do nothing if already exists
        pass
    mdo._raw_ontology = mdo._get_ontology()
    mdo._raw_ontology.append(FAKE_NUMBER_FIELD)
    mdo._build_ontology()
    yield mdo


@pytest.fixture
def big_dataset(dataset: Dataset, image_url):
    task = dataset.create_data_rows(
        [
            {"row_data": image_url, "external_id": "my-image"},
        ]
        * 5
    )
    task.wait_till_done()

    yield dataset


def make_metadata(dr_id: str = None, gk: str = None) -> DataRowMetadata:
    msg = "A message"
    time = datetime.now(timezone.utc)

    metadata = DataRowMetadata(
        global_key=gk,
        data_row_id=dr_id,
        fields=[
            DataRowMetadataField(
                schema_id=SPLIT_SCHEMA_ID, value=TEST_SPLIT_ID
            ),
            DataRowMetadataField(schema_id=CAPTURE_DT_SCHEMA_ID, value=time),
            DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value=msg),
        ],
    )
    return metadata


def make_named_metadata(dr_id) -> DataRowMetadata:
    msg = "A message"
    time = datetime.now(timezone.utc)

    metadata = DataRowMetadata(
        data_row_id=dr_id,
        fields=[
            DataRowMetadataField(name="split", value=TEST_SPLIT_ID),
            DataRowMetadataField(name="captureDateTime", value=time),
            DataRowMetadataField(name=CUSTOM_TEXT_SCHEMA_NAME, value=msg),
        ],
    )
    return metadata


def test_bulk_export_datarow_metadata(data_row, mdo: DataRowMetadataOntology):
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    exported = mdo.bulk_export([data_row.uid])
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3

    exported = mdo.bulk_export(UniqueIds([data_row.uid]))
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3

    exported = mdo.bulk_export(GlobalKeys([data_row.global_key]))
    assert exported[0].global_key == data_row.global_key
    assert exported[0].data_row_id == data_row.uid
    assert len([field for field in exported[0].fields]) == 3


def test_get_datarow_metadata_ontology(mdo):
    assert len(mdo.fields)
    assert len(mdo.reserved_fields)
    # two are created by mdo fixture but there may be more
    assert len(mdo.custom_fields) >= 2

    split = mdo.reserved_by_name["split"]["train"]

    assert DataRowMetadata(
        data_row_id=FAKE_DATAROW_ID,
        fields=[
            DataRowMetadataField(
                schema_id=mdo.reserved_by_name["captureDateTime"].uid,
                value=datetime.now(timezone.utc),
            ),
            DataRowMetadataField(schema_id=split.parent, value=split.uid),
            DataRowMetadataField(
                schema_id=mdo.reserved_by_name["tag"].uid, value="hello-world"
            ),
        ],
    )


def test_bulk_upsert_datarow_metadata(data_row, mdo: DataRowMetadataOntology):
    metadata = make_metadata(data_row.uid)
    mdo.bulk_upsert([metadata])
    exported = mdo.bulk_export([data_row.uid])
    assert len(exported)
    assert len([field for field in exported[0].fields]) == 3


def test_bulk_upsert_datarow_metadata_by_globalkey(
    data_rows, mdo: DataRowMetadataOntology
):
    global_keys = [data_row.global_key for data_row in data_rows]
    metadata = [make_metadata(gk=global_key) for global_key in global_keys]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0


@pytest.mark.slow
def test_large_bulk_upsert_datarow_metadata(big_dataset, mdo):
    metadata = []
    data_row_ids = [dr.uid for dr in big_dataset.data_rows()]
    for data_row_id in data_row_ids:
        metadata.append(make_metadata(data_row_id))
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata_lookup = {
        metadata.data_row_id: metadata
        for metadata in mdo.bulk_export(data_row_ids)
    }
    for data_row_id in data_row_ids:
        assert len(
            [f for f in metadata_lookup.get(data_row_id).fields]
        ), metadata_lookup.get(data_row_id).fields


def test_upsert_datarow_metadata_by_name(data_row, mdo):
    metadata = [make_named_metadata(data_row.uid)]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata_lookup = {
        metadata.data_row_id: metadata
        for metadata in mdo.bulk_export([data_row.uid])
    }
    assert len(
        [f for f in metadata_lookup.get(data_row.uid).fields]
    ), metadata_lookup.get(data_row.uid).fields


def test_upsert_datarow_metadata_option_by_name(data_row, mdo):
    metadata = DataRowMetadata(
        data_row_id=data_row.uid,
        fields=[
            DataRowMetadataField(name="split", value="test"),
        ],
    )
    errors = mdo.bulk_upsert([metadata])
    assert len(errors) == 0

    datarows = mdo.bulk_export([data_row.uid])
    assert len(datarows[0].fields) == 1
    metadata = datarows[0].fields[0]
    assert metadata.schema_id == SPLIT_SCHEMA_ID
    assert metadata.name == "test"
    assert metadata.value == TEST_SPLIT_ID


def test_upsert_datarow_metadata_option_by_incorrect_name(data_row, mdo):
    metadata = DataRowMetadata(
        data_row_id=data_row.uid,
        fields=[
            DataRowMetadataField(name="split", value="test1"),
        ],
    )
    with pytest.raises(KeyError):
        mdo.bulk_upsert([metadata])


def test_raise_enum_upsert_schema_error(data_row, mdo):
    """Setting an option id as the schema id will raise a Value Error"""

    metadata = DataRowMetadata(
        data_row_id=data_row.uid,
        fields=[
            DataRowMetadataField(
                schema_id=TEST_SPLIT_ID, value=SPLIT_SCHEMA_ID
            ),
        ],
    )
    with pytest.raises(ValueError):
        mdo.bulk_upsert([metadata])


def test_upsert_non_existent_schema_id(data_row, mdo):
    """Raise error on non-existent schema id"""
    metadata = DataRowMetadata(
        data_row_id=data_row.uid,
        fields=[
            DataRowMetadataField(schema_id=INVALID_SCHEMA_ID, value="message"),
        ],
    )
    with pytest.raises(ValueError):
        mdo.bulk_upsert([metadata])


def test_parse_raw_metadata(mdo):
    example = {
        "dataRowId": "ckr6kkfx801ui0yrtg9fje8xh",
        "globalKey": "global-key-1",
        "fields": [
            {
                "schemaId": "cko8s9r5v0001h2dk9elqdidh",
                "value": "my-new-message",
            },
            {"schemaId": "cko8sbczn0002h2dkdaxb5kal", "value": {}},
            {"schemaId": "cko8sbscr0003h2dk04w86hof", "value": {}},
            {
                "schemaId": "cko8sdzv70006h2dk8jg64zvb",
                "value": "2021-07-20T21:41:14.606710Z",
            },
            {"schemaId": FAKE_SCHEMA_ID, "value": 0.5},
        ],
    }

    parsed = mdo.parse_metadata([example])
    assert len(parsed) == 1
    for row in parsed:
        assert row.data_row_id == example["dataRowId"]
        assert row.global_key == example["globalKey"]
        assert len(row.fields) == 4

    for row in parsed:
        for field in row.fields:
            assert mdo._parse_upsert(field)


def test_parse_raw_metadata_fields(mdo):
    example = [
        {"schemaId": "cko8s9r5v0001h2dk9elqdidh", "value": "my-new-message"},
        {"schemaId": "cko8sbczn0002h2dkdaxb5kal", "value": {}},
        {"schemaId": "cko8sbscr0003h2dk04w86hof", "value": {}},
        {
            "schemaId": "cko8sdzv70006h2dk8jg64zvb",
            "value": "2021-07-20T21:41:14.606710Z",
        },
        {"schemaId": FAKE_SCHEMA_ID, "value": 0.5},
    ]

    parsed = mdo.parse_metadata_fields(example)
    assert len(parsed) == 4

    for field in parsed:
        assert mdo._parse_upsert(field)


def test_parse_metadata_schema():
    unparsed = {
        "id": "cl467a4ec0046076g7s9yheoa",
        "name": "enum metadata",
        "kind": "CustomMetadataEnum",
        "options": [
            {
                "id": "cl467a4ec0047076ggjneeruy",
                "name": "option1",
                "kind": "CustomMetadataEnumOption",
            },
            {
                "id": "cl4qa31u0009e078p5m280jer",
                "name": "option2",
                "kind": "CustomMetadataEnumOption",
            },
        ],
    }
    parsed = _parse_metadata_schema(unparsed)
    assert parsed.uid == "cl467a4ec0046076g7s9yheoa"
    assert parsed.name == "enum metadata"
    assert parsed.kind == DataRowMetadataKind.enum
    assert len(parsed.options) == 2
    assert parsed.options[0].uid == "cl467a4ec0047076ggjneeruy"
    assert parsed.options[0].kind == DataRowMetadataKind.option


def test_create_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(
        metadata_name, DataRowMetadataKind.enum, ["option 1", "option 2"]
    )
    assert created_schema.name == metadata_name
    assert created_schema.kind == DataRowMetadataKind.enum
    assert len(created_schema.options) == 2
    assert created_schema.options[0].name == "option 1"
    mdo.delete_schema(metadata_name)


def test_update_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(
        metadata_name, DataRowMetadataKind.enum, ["option 1", "option 2"]
    )
    updated_schema = mdo.update_schema(
        metadata_name, f"{metadata_name}_updated"
    )
    assert updated_schema.name == f"{metadata_name}_updated"
    assert updated_schema.uid == created_schema.uid
    assert updated_schema.kind == DataRowMetadataKind.enum
    mdo.delete_schema(f"{metadata_name}_updated")


def test_update_enum_options(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(
        metadata_name, DataRowMetadataKind.enum, ["option 1", "option 2"]
    )
    updated_schema = mdo.update_enum_option(
        metadata_name, "option 1", "option 3"
    )
    assert updated_schema.name == metadata_name
    assert updated_schema.uid == created_schema.uid
    assert updated_schema.kind == DataRowMetadataKind.enum
    assert updated_schema.options[0].uid == created_schema.options[0].uid
    assert updated_schema.options[0].name == "option 3"
    mdo.delete_schema(metadata_name)


def test_delete_schema(mdo):
    metadata_name = str(uuid.uuid4())
    created_schema = mdo.create_schema(
        metadata_name, DataRowMetadataKind.string
    )
    status = mdo.delete_schema(created_schema.name)
    mdo.refresh_ontology()
    assert status
    assert metadata_name not in mdo.custom_by_name


@pytest.mark.parametrize(
    "datetime_str", ["2011-11-04T00:05:23Z", "2011-11-04T00:05:23+00:00"]
)
def test_upsert_datarow_date_metadata(data_row, mdo, datetime_str):
    metadata = [
        DataRowMetadata(
            data_row_id=data_row.uid,
            fields=[
                DataRowMetadataField(
                    name="captureDateTime", value=datetime_str
                ),
            ],
        )
    ]
    errors = mdo.bulk_upsert(metadata)
    assert len(errors) == 0

    metadata = mdo.bulk_export([data_row.uid])
    assert f"{metadata[0].fields[0].value}" == "2011-11-04 00:05:23+00:00"


@pytest.mark.parametrize(
    "datetime_str", ["2011-11-04T00:05:23Z", "2011-11-04T00:05:23+00:00"]
)
def test_create_data_row_with_metadata(dataset, image_url, datetime_str):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    metadata_fields = [
        DataRowMetadataField(name="captureDateTime", value=datetime_str)
    ]

    data_row = dataset.create_data_row(
        row_data=image_url, metadata_fields=metadata_fields
    )

    retrieved_data_row = client.get_data_row(data_row.uid)
    assert (
        f"{retrieved_data_row.metadata[0].value}" == "2011-11-04 00:05:23+00:00"
    )



from collections import defaultdict
from enum import Enum
from typing import Iterator, Set

from lbox.exceptions import (
    MalformedQueryException,
    ResourceCreationError,
    ResourceNotFoundError,
    UnprocessableEntityError,
)
from pydantic import BaseModel, ConfigDict

from labelbox import Client
from labelbox.schema.media_type import MediaType
from labelbox.schema.ontology_kind import EditorTaskType
from labelbox.schema.project import Project
from labelbox.schema.user import User


class UserGroupColor(Enum):
    """
    Enum representing the colors available for a group.

    Attributes:
        BLUE (str): Hex color code for blue (#9EC5FF).
        PURPLE (str): Hex color code for purple (#CEB8FF).
        ORANGE (str): Hex color code for orange (#FFB35F).
        CYAN (str): Hex color code for cyan (#4ED2F9).
        PINK (str): Hex color code for pink (#FFAEA9).
        LIGHT_PINK (str): Hex color code for light pink (#FFA9D5).
        GREEN (str): Hex color code for green (#3FDC9A).
        YELLOW (str): Hex color code for yellow (#E7BF00).
        GRAY (str): Hex color code for gray (#B8C4D3).
    """

    BLUE = "9EC5FF"
    PURPLE = "CEB8FF"
    ORANGE = "FFB35F"
    CYAN = "4ED2F9"
    PINK = "FFAEA9"
    LIGHT_PINK = "FFA9D5"
    GREEN = "3FDC9A"
    YELLOW = "E7BF00"
    GRAY = "B8C4D3"


class UserGroup(BaseModel):
    """
    Represents a user group in Labelbox.

    Attributes:
        id (str): The ID of the user group.
        name (str): The name of the user group.
        color (UserGroupColor): The color of the user group.
        users (Set[UserGroupUser]): The set of users in the user group.
        projects (Set[UserGroupProject]): The set of projects associated with the user group.
        client (Client): The Labelbox client object.

    Methods:
        __init__(self, client: Client)
        get(self) -> "UserGroup"
        update(self) -> "UserGroup"
        create(self) -> "UserGroup"
        delete(self) -> bool
        get_user_groups(client: Client) -> Iterator["UserGroup"]
    """

    id: str
    name: str
    color: UserGroupColor
    users: Set[User]
    projects: Set[Project]
    client: Client
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        client: Client,
        id: str = "",
        name: str = "",
        color: UserGroupColor = UserGroupColor.BLUE,
        users: Set[User] = set(),
        projects: Set[Project] = set(),
    ):
        """
        Initializes a UserGroup object.

        Args:
            client (Client): The Labelbox client object.
            id (str, optional): The ID of the user group. Defaults to an empty string.
            name (str, optional): The name of the user group. Defaults to an empty string.
            color (UserGroupColor, optional): The color of the user group. Defaults to UserGroupColor.BLUE.
            users (Set[User], optional): The set of users in the user group. Defaults to an empty set.
            projects (Set[Project], optional): The set of projects associated with the user group. Defaults to an empty set.
        """
        super().__init__(
            client=client,
            id=id,
            name=name,
            color=color,
            users=users,
            projects=projects,
        )

    def get(self) -> "UserGroup":
        """
        Reloads the user group information from the server.

        This method sends a GraphQL query to the server to fetch the latest information
        about the user group, including its name, color, projects, and members. The fetched
        information is then used to update the corresponding attributes of the `Group` object.

        Returns:
            UserGroup: The updated `UserGroup` object.

        Raises:
            ResourceNotFoundError: If the query fails to fetch the group information.
            ValueError: If the group ID is not provided.
        """
        if not self.id:
            raise ValueError("Group id is required")
        query = """
            query GetUserGroupPyApi($id: ID!) {
                userGroup(where: {id: $id}) {
                    id
                    name
                    color
                    projects {
                        nodes {
                            id
                            name
                        }
                        totalCount
                    }
                    members {
                        nodes {
                            id
                            email
                        }
                        totalCount
                    }
                }
            }        
        """
        params = {
            "id": self.id,
        }
        result = self.client.execute(query, params)
        if not result:
            raise ResourceNotFoundError(
                message="Failed to get user group as user group does not exist"
            )
        self.name = result["userGroup"]["name"]
        self.color = UserGroupColor(result["userGroup"]["color"])
        self.projects = self._get_projects_set(
            result["userGroup"]["projects"]["nodes"]
        )
        self.users = self._get_users_set(
            result["userGroup"]["members"]["nodes"]
        )
        return self

    def update(self) -> "UserGroup":
        """
        Updates the group in Labelbox.

        Returns:
            UserGroup: The updated UserGroup object. (self)

        Raises:
            ResourceNotFoundError: If the update fails due to unknown user group
            UnprocessableEntityError: If the update fails due to a malformed input
            ValueError: If the group id or name is not provided
        """
        if not self.id:
            raise ValueError("Group id is required")
        if not self.name:
            raise ValueError("Group name is required")
        query = """
        mutation UpdateUserGroupPyApi($id: ID!, $name: String!, $color: String!, $projectIds: [String!]!, $userIds: [String!]!) {
            updateUserGroup(
                where: {id: $id}
                data: {name: $name, color: $color, projectIds: $projectIds, userIds: $userIds}
            ) {
                group {
                    id
                    name
                    color
                    projects {
                        nodes {
                            id
                            name
                        }
                    }
                    members {
                        nodes {
                            id
                            email
                        }
                    }
                }
            }
        }
        """
        params = {
            "id": self.id,
            "name": self.name,
            "color": self.color.value,
            "projectIds": [project.uid for project in self.projects],
            "userIds": [user.uid for user in self.users],
        }
        try:
            result = self.client.execute(query, params)
            if not result:
                raise ResourceNotFoundError(
                    message="Failed to update user group as user group does not exist"
                )
        except MalformedQueryException as e:
            raise UnprocessableEntityError("Failed to update user group") from e
        return self

    def create(self) -> "UserGroup":
        """
        Creates a new user group.

        Raises:
            ResourceCreationError: If the group already exists.
            ValueError: If the group name is not provided.

        Returns:
            UserGroup: The created user group.
        """
        if self.id:
            raise ResourceCreationError("Group already exists")
        if not self.name:
            raise ValueError("Group name is required")
        query = """
        mutation CreateUserGroupPyApi($name: String!, $color: String!, $projectIds: [String!]!, $userIds: [String!]!) {
            createUserGroup(
                data: {
                    name: $name,
                    color: $color,
                    projectIds: $projectIds,
                    userIds: $userIds
                }
            ) {
                group {
                    id
                    name
                    color
                    projects {
                        nodes {
                            id
                            name
                        }
                    }
                    members {
                        nodes {
                            id
                            email
                        }
                    }
                }
            }
        }
        """
        params = {
            "name": self.name,
            "color": self.color.value,
            "projectIds": [project.uid for project in self.projects],
            "userIds": [user.uid for user in self.users],
        }
        result = None
        error = None
        try:
            result = self.client.execute(query, params)
        except Exception as e:
            error = e
        if not result or error:
            # This is client side only, server doesn't have an equivalent error
            raise ResourceCreationError(
                f"Failed to create user group, either user group name is in use currently, or provided user or projects don't exist server error: {error}"
            )
        result = result["createUserGroup"]["group"]
        self.id = result["id"]
        return self

    def delete(self) -> bool:
        """
        Deletes the user group from Labelbox.

        This method sends a mutation request to the Labelbox API to delete the user group
        with the specified ID. If the deletion is successful, it returns True. Otherwise,
        it raises an UnprocessableEntityError and returns False.

        Returns:
            bool: True if the user group was successfully deleted, False otherwise.

        Raises:
            ResourceNotFoundError: If the deletion of the user group fails due to not existing
            ValueError: If the group ID is not provided.
        """
        if not self.id:
            raise ValueError("Group id is required")
        query = """
        mutation DeleteUserGroupPyApi($id: ID!) {
            deleteUserGroup(where: {id: $id}) {
                success
            }
        }
        """
        params = {"id": self.id}
        result = self.client.execute(query, params)
        if not result:
            raise ResourceNotFoundError(
                message="Failed to delete user group as user group does not exist"
            )
        return result["deleteUserGroup"]["success"]

    def get_user_groups(self) -> Iterator["UserGroup"]:
        """
        Gets all user groups in Labelbox.

        Args:
            client (Client): The Labelbox client.

        Returns:
            Iterator[UserGroup]: An iterator over the user groups.
        """
        query = """
            query GetUserGroupsPyApi($after: String) {
                userGroups(after: $after) {
                    nodes {
                        id
                        name
                        color
                        projects {
                            nodes {
                                id
                                name
                            }
                            totalCount
                        }
                        members {
                            nodes {
                                id
                                email
                            }
                            totalCount
                        }
                    }
                    nextCursor
                }
            }
        """
        nextCursor = None
        while True:
            userGroups = self.client.execute(query, {"after": nextCursor})[
                "userGroups"
            ]
            if not userGroups:
                return
                yield
            groups = userGroups["nodes"]
            for group in groups:
                userGroup = UserGroup(self.client)
                userGroup.id = group["id"]
                userGroup.name = group["name"]
                userGroup.color = UserGroupColor(group["color"])
                userGroup.users = self._get_users_set(group["members"]["nodes"])
                userGroup.projects = self._get_projects_set(
                    group["projects"]["nodes"]
                )
                yield userGroup
            nextCursor = userGroups["nextCursor"]
            if not nextCursor:
                return
                yield

    def _get_users_set(self, user_nodes):
        """
        Retrieves a set of User objects from the given user nodes.

        Args:
            user_nodes (list): A list of user nodes containing user information.

        Returns:
            set: A set of User objects.
        """
        users = set()
        for user in user_nodes:
            user_values = defaultdict(lambda: None)
            user_values["id"] = user["id"]
            user_values["email"] = user["email"]
            users.add(User(self.client, user_values))
        return users

    def _get_projects_set(self, project_nodes):
        """
        Retrieves a set of projects based on the given project nodes.

        Args:
            project_nodes (list): A list of project nodes.

        Returns:
            set: A set of Project objects.
        """
        projects = set()
        for project in project_nodes:
            project_values = defaultdict(lambda: None)
            project_values["id"] = project["id"]
            project_values["name"] = project["name"]
            project_values["editorTaskType"] = EditorTaskType.Missing.value
            project_values["mediaType"] = MediaType.Image.value
            projects.add(Project(self.client, project_values))
        return projects



from typing import cast, Any, Dict, List, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from labelbox.types import Label


def serialize_labels(
    objects: Union[List[Dict[str, Any]], List["Label"]],
) -> List[Dict[str, Any]]:
    """
    Checks if objects are of type Labels and serializes labels for annotation import. Serialization depends the labelbox[data] package, therefore NDJsonConverter is only loaded if using `Label` objects instead of `dict` objects.
    """
    if len(objects) == 0:
        return []

    is_label_type = not isinstance(objects[0], Dict)
    if is_label_type:
        # If a Label object exists, labelbox[data] is already installed, so no error checking is needed.
        from labelbox.data.serialization import NDJsonConverter

        labels = cast(List["Label"], objects)
        return list(NDJsonConverter.serialize(labels))

    return cast(List[Dict[str, Any]], objects)
